# encoding: utf-8

# -------------------------------------------------------------------------------
# (C) Copyright 2019 - 2021 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates
# and licensors (“Micro Focus”) are as may be set forth in the express warranty
# statements accompanying such products and services. Nothing herein should be
# construed as constituting an additional warranty. Micro Focus shall not be
# liable for technical or editorial errors or omissions contained herein. The
# information contained herein is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains confidential
# information and a valid license is required for possession, use or copying. If
# this work is provided to the U.S. Government, consistent with FAR 12.211 and
# 12.212, Commercial Computer Software, Computer Software Documentation, and
# Technical Data for Commercial Items  are licensed to the U.S. Government under
# vendor's standard commercial license.
# -------------------------------------------------------------------------------

# File:	clients/create-answer-file.ycp
# Package:	Answer file generator for Channel Upgrade Configuration of OES 2 products
# Summary:	Client for generating the "answer" file
# Authors:	Mark Clark <mclark@novell.com>
#
# $Id: create-answer-file.ycp,v 1.0.0.0 2009/05/13 10:54:44 mclark Exp $
#
# This is a client that generates the "answer" file for the Channel upgrade option of OES 11. It takes one
# parameter, the password, and performs a simple obfuscation and writes it to the answer file.  The
# channel-upgrade script will unencrypt the password when required.
module Yast
  class CreateAnswerFileClient < Client
    def main
      Yast.import "UI"

      textdomain "novell-base"

      Yast.import "Label"
      Yast.import "Wizard"
      Yast.import "Report"
      Yast.import "Popup"
      # Reading passwords from command line
      @clear_password = Convert.to_string(WFM.Args(0))
      @dsfw_clear_password = Convert.to_string(WFM.Args(1))

      # if command line parameter is help/--help, display help
      if @clear_password == "help"
        @help = _(
          "Use any of the following methods to generate an answer key file.\n" +
            "\n" +
            "1. Direct Generation\n" +
            "yast2 /usr/share/yast2/clients/create-answer-file.ycp <eDirectory Admin Password> <DSfW Admin Password>\n" +
            "Note: This method is not recommended as the passwords are stored in the y2log in clear text.\n" +
            "\n" +
            "2. Exporting the Passwords to Variables\n" +
            "export OES_EDIR_DATA=<eDirectory Admin Password>\n" +
            "export OES_DSFW_DATA=<DSfW Admin Password>\n" +
            "yast2 /usr/share/yast2/clients/create-answer-file.ycp\n" +
            "\n" +
            "3. Using the GUI\n" +
            "yast2 /usr/share/yast2/clients/create-answer-file.ycp\n" +
            "In the YaST2 dialog, specify the eDirectory and DSfW passwords, then click OK.\n" +
            "Note: DSfW password should be specified only if you are upgrading a DSfW server."
        )

        Popup.Message(@help)

        return
      end
      if @clear_password == "" || @clear_password == nil
        # Reading password from environmental varibles
        @clear_password = Builtins.getenv("OES_EDIR_DATA")
        @dsfw_clear_password = Builtins.getenv("OES_DSFW_DATA")

        # if environment variables are not set, popup widget
        if @clear_password == "" || @clear_password == nil
          @caption = _("Answer file creation")
          UI.OpenDialog(
            Opt(:decorated),
            VBox(
              HSpacing(1),
              VBox(
                Label(_("Answer file creation")),
                VSpacing(1),
                TextEntry(
                  Id(:clear_password),
                  _("eDirectory Admin Password"),
                  @clear_password
                ),
                Left(
                  CheckBox(
                    Id(:dsfw_password_required),
                    Opt(:notify),
                    _("If DSfW is configured"),
                    false
                  )
                ),
                TextEntry(
                  Id(:dsfw_clear_password),
                  _("DSfW Admin Password"),
                  @dsfw_clear_password
                )
              ),
              HSpacing(1),
              HBox(
                PushButton(Id(:ok), Opt(:default), Label.OKButton),
                PushButton(Id(:cancel), Label.CancelButton),
                PushButton(Id(:help), Label.HelpButton)
              )
            )
          )

          UI.ChangeWidget(Id(:dsfw_clear_password), :Enabled, false)
          @ret = nil
          while true
            @ret = UI.UserInput
            # abort?
            if @ret == :cancel
              UI.CloseDialog
              break
            elsif @ret == :dsfw_password_required
              @dsfw_password_required = Convert.to_boolean(
                UI.QueryWidget(Id(:dsfw_password_required), :Value)
              )
              if @dsfw_password_required
                UI.ChangeWidget(Id(:dsfw_clear_password), :Enabled, true)
              else
                UI.ChangeWidget(Id(:dsfw_clear_password), :Enabled, false)
                UI.ChangeWidget(Id(:dsfw_clear_password), :Value, "")
              end
            elsif @ret == :help
              @help = _(
                "Specify the eDirectory and DSfW administrator passwords in the respective fields.\nDSfW password should be specified only in case of a DSfW server upgrade."
              )
              Popup.Message(@help)
            elsif @ret == :ok
              @clear_password = Convert.to_string(
                UI.QueryWidget(Id(:clear_password), :Value)
              )
              @dsfw_clear_password = Convert.to_string(
                UI.QueryWidget(Id(:dsfw_clear_password), :Value)
              )
              if @clear_password == ""
                @err = _("Admin Password cannot be empty.")
                Report.Error(@err)
                next
              end
              # Make sure the password only contains valid characters
              next if ValidatePassword(@clear_password) == false

              if @dsfw_clear_password != nil && @dsfw_clear_password != ""
                next if ValidatePassword(@dsfw_clear_password) == false
              end

              # Obfuscate the password
              @obf_password = Obfuscate(@clear_password)
              if @dsfw_clear_password != nil && @dsfw_clear_password != ""
                @dsfw_obf_password = Obfuscate(@dsfw_clear_password)
              end

              # Write "answer" file with obfuscated passwords
              if @obf_password != nil && @obf_password != ""
                @writeStr = @obf_password

                # Start out with a blank file.
                @ret2 = SCR.Write(path(".target.string"), "answer", "")

                if @dsfw_obf_password != nil && @dsfw_obf_password != ""
                  @writeStr = Ops.add(
                    Ops.add(@writeStr, "\n----\n"),
                    @dsfw_obf_password
                  )
                end
                # Write the password data
                @ret2 = SCR.Write(
                  path(".target.string"),
                  "answer",
                  Ops.add(@writeStr, "\n")
                )

                # Flush the cache
                @ret2 = SCR.Write(path(".target.string"), "answer", nil)

                @clearPass = UnObfuscate(@obf_password) # Get the first password
                @msg = Builtins.sformat(
                  _("Decrypted password (sanity check) is '%1'. Size is %2"),
                  @clearPass,
                  Builtins.size(@clearPass)
                )

                if @dsfw_obf_password != nil && @dsfw_obf_password != ""
                  @dsfwClearPass = UnObfuscate(@dsfw_obf_password) # Get the dsfw password
                  @msg = Builtins.sformat(
                    _(
                      "Decrypted passwords (sanity check) are '%1', size is %2, '%3', size is %4"
                    ),
                    @clearPass,
                    Builtins.size(@clearPass),
                    @dsfwClearPass,
                    Builtins.size(@dsfwClearPass)
                  )
                end

                Popup.Message(@msg)
                UI.CloseDialog
                break
              end

              Builtins.y2milestone("ChannelUpgrade:create-answer-file finished")
              Builtins.y2milestone("-------------------------------")
            else
              Builtins.y2error("unexpected retcode: %1", @ret)
              next
            end
          end
        else
          validate_password(@clear_password, @dsfw_clear_password)
        end
      else
        validate_password(@clear_password, @dsfw_clear_password)
      end

      nil
    end

    # Assign a random number to each valid password character
    def LoadTranslationMap
      ret = {}

      Ops.set(ret, "-", 46) # 1
      Ops.set(ret, "0", 122) # 2
      Ops.set(ret, "1", 18) # 3
      Ops.set(ret, "2", 6) # 4
      Ops.set(ret, "3", 20) # 5
      Ops.set(ret, "4", 32) # 6
      Ops.set(ret, "5", 10) # 7
      Ops.set(ret, "6", 60) # 8
      Ops.set(ret, "7", 70) # 9
      Ops.set(ret, "8", 5) # 10
      Ops.set(ret, "9", 66) # 11
      Ops.set(ret, "a", 209) # 12
      Ops.set(ret, "b", 11) # 13
      Ops.set(ret, "c", 168) # 14
      Ops.set(ret, "d", 102) # 15
      Ops.set(ret, "e", 180) # 16
      Ops.set(ret, "f", 2) # 17
      Ops.set(ret, "g", 152) # 18
      Ops.set(ret, "h", 80) # 19
      Ops.set(ret, "i", 162) # 20
      Ops.set(ret, "j", 187) # 21
      Ops.set(ret, "k", 15) # 22
      Ops.set(ret, "l", 151) # 23
      Ops.set(ret, "m", 21) # 24
      Ops.set(ret, "n", 87) # 25
      Ops.set(ret, "o", 30) # 26
      Ops.set(ret, "p", 1) # 27
      Ops.set(ret, "q", 65) # 28
      Ops.set(ret, "r", 163) # 29
      Ops.set(ret, "s", 81) # 30
      Ops.set(ret, "t", 67) # 31
      Ops.set(ret, "u", 71) # 32
      Ops.set(ret, "v", 17) # 33
      Ops.set(ret, "w", 167) # 34
      Ops.set(ret, "x", 25) # 35
      Ops.set(ret, "y", 185) # 36
      Ops.set(ret, "z", 104) # 37
      Ops.set(ret, "A", 24) # 38
      Ops.set(ret, "B", 61) # 39
      Ops.set(ret, "C", 88) # 40
      Ops.set(ret, "D", 186) # 41
      Ops.set(ret, "E", 33) # 42
      Ops.set(ret, "F", 14) # 43
      Ops.set(ret, "G", 166) # 44
      Ops.set(ret, "H", 72) # 45
      Ops.set(ret, "I", 85) # 46
      Ops.set(ret, "J", 62) # 47
      Ops.set(ret, "K", 36) # 48
      Ops.set(ret, "L", 106) # 49
      Ops.set(ret, "M", 26) # 50
      Ops.set(ret, "N", 181) # 51
      Ops.set(ret, "O", 206) # 52
      Ops.set(ret, "P", 207) # 53
      Ops.set(ret, "Q", 3) # 54
      Ops.set(ret, "R", 74) # 55
      Ops.set(ret, "S", 22) # 56
      Ops.set(ret, "T", 164) # 57
      Ops.set(ret, "U", 86) # 58
      Ops.set(ret, "V", 68) # 59
      Ops.set(ret, "W", 183) # 60
      Ops.set(ret, "X", 103) # 61
      Ops.set(ret, "Y", 35) # 62
      Ops.set(ret, "Z", 150) # 63
      Ops.set(ret, "!", 82) # 64
      Ops.set(ret, "@", 160) # 65
      Ops.set(ret, "#", 31) # 66
      Ops.set(ret, "$", 16) # 67
      Ops.set(ret, "%", 182) # 68
      Ops.set(ret, "^", 63) # 69
      Ops.set(ret, "&", 84) # 70
      Ops.set(ret, "*", 165) # 71
      Ops.set(ret, "~", 23) # 72
      Ops.set(ret, "(", 184) # 73
      Ops.set(ret, ")", 101) # 74
      Ops.set(ret, ",", 161) # 75
      Ops.set(ret, ";", 89) # 76
      Ops.set(ret, ":", 205) # 77
      Ops.set(ret, ".", 64) # 78
      Ops.set(ret, "_", 13) # 79
      Ops.set(ret, "+", 34) # 80
      Ops.set(ret, "/", 83) # 81
      Ops.set(ret, "|", 69) # 82
      Ops.set(ret, "?", 73) # 83
      Ops.set(ret, "{", 4) # 84
      Ops.set(ret, "}", 208) # 85
      Ops.set(ret, "=", 100) # 86
      Ops.set(ret, "[", 105) # 87
      Ops.set(ret, "]", 12) # 88

      deep_copy(ret)
    end

    # Assign a valid password character to a random number
    def LoadDeTranslationMap
      ret = {}

      Ops.set(ret, 46, "-") # 1
      Ops.set(ret, 122, "0") # 2
      Ops.set(ret, 18, "1") # 3
      Ops.set(ret, 6, "2") # 4
      Ops.set(ret, 20, "3") # 5
      Ops.set(ret, 32, "4") # 6
      Ops.set(ret, 10, "5") # 7
      Ops.set(ret, 60, "6") # 8
      Ops.set(ret, 70, "7") # 9
      Ops.set(ret, 5, "8") # 10
      Ops.set(ret, 66, "9") # 11
      Ops.set(ret, 209, "a") # 12
      Ops.set(ret, 11, "b") # 13
      Ops.set(ret, 168, "c") # 14
      Ops.set(ret, 102, "d") # 15
      Ops.set(ret, 180, "e") # 16
      Ops.set(ret, 2, "f") # 17
      Ops.set(ret, 152, "g") # 18
      Ops.set(ret, 80, "h") # 19
      Ops.set(ret, 162, "i") # 20
      Ops.set(ret, 187, "j") # 21
      Ops.set(ret, 15, "k") # 22
      Ops.set(ret, 151, "l") # 23
      Ops.set(ret, 21, "m") # 24
      Ops.set(ret, 87, "n") # 25
      Ops.set(ret, 30, "o") # 26
      Ops.set(ret, 1, "p") # 27
      Ops.set(ret, 65, "q") # 28
      Ops.set(ret, 163, "r") # 29
      Ops.set(ret, 81, "s") # 30
      Ops.set(ret, 67, "t") # 31
      Ops.set(ret, 71, "u") # 32
      Ops.set(ret, 17, "v") # 33
      Ops.set(ret, 167, "w") # 34
      Ops.set(ret, 25, "x") # 35
      Ops.set(ret, 185, "y") # 36
      Ops.set(ret, 104, "z") # 37
      Ops.set(ret, 24, "A") # 38
      Ops.set(ret, 61, "B") # 39
      Ops.set(ret, 88, "C") # 40
      Ops.set(ret, 186, "D") # 41
      Ops.set(ret, 33, "E") # 42
      Ops.set(ret, 14, "F") # 43
      Ops.set(ret, 166, "G") # 44
      Ops.set(ret, 72, "H") # 45
      Ops.set(ret, 85, "I") # 46
      Ops.set(ret, 62, "J") # 47
      Ops.set(ret, 36, "K") # 48
      Ops.set(ret, 106, "L") # 49
      Ops.set(ret, 26, "M") # 50
      Ops.set(ret, 181, "N") # 51
      Ops.set(ret, 206, "O") # 52
      Ops.set(ret, 207, "P") # 53
      Ops.set(ret, 3, "Q") # 54
      Ops.set(ret, 74, "R") # 55
      Ops.set(ret, 22, "S") # 56
      Ops.set(ret, 164, "T") # 57
      Ops.set(ret, 86, "U") # 58
      Ops.set(ret, 68, "V") # 59
      Ops.set(ret, 183, "W") # 60
      Ops.set(ret, 103, "X") # 61
      Ops.set(ret, 35, "Y") # 62
      Ops.set(ret, 150, "Z") # 63
      Ops.set(ret, 82, "!") # 64
      Ops.set(ret, 160, "@") # 65
      Ops.set(ret, 31, "#") # 66
      Ops.set(ret, 16, "$") # 67
      Ops.set(ret, 182, "%") # 68
      Ops.set(ret, 63, "^") # 69
      Ops.set(ret, 84, "&") # 70
      Ops.set(ret, 165, "*") # 71
      Ops.set(ret, 23, "~") # 72
      Ops.set(ret, 184, "(") # 73
      Ops.set(ret, 101, ")") # 74
      Ops.set(ret, 161, ",") # 75
      Ops.set(ret, 89, ";") # 76
      Ops.set(ret, 205, ":") # 77
      Ops.set(ret, 64, ".") # 78
      Ops.set(ret, 13, "_") # 79
      Ops.set(ret, 34, "+") # 80
      Ops.set(ret, 83, "/") # 81
      Ops.set(ret, 69, "|") # 82
      Ops.set(ret, 73, "?") # 83
      Ops.set(ret, 4, "{") # 84
      Ops.set(ret, 208, "}") # 85
      Ops.set(ret, 100, "=") # 86
      Ops.set(ret, 105, "[") # 87
      Ops.set(ret, 12, "]") # 88

      deep_copy(ret)
    end

    # How much to bit shift the value based on string position
    def LoadBitshiftList
      ret = []

      Ops.set(ret, 0, 3) # 1
      Ops.set(ret, 1, 1) # 2
      Ops.set(ret, 2, 2) # 3
      Ops.set(ret, 3, 4) # 4
      Ops.set(ret, 4, 3) # 5
      Ops.set(ret, 5, 1) # 6
      Ops.set(ret, 6, 4) # 7
      Ops.set(ret, 7, 2) # 8
      Ops.set(ret, 8, 4) # 9
      Ops.set(ret, 9, 2) # 10
      Ops.set(ret, 10, 1) # 11
      Ops.set(ret, 11, 3) # 12
      Ops.set(ret, 12, 1) # 13
      Ops.set(ret, 13, 2) # 14
      Ops.set(ret, 14, 1) # 15
      Ops.set(ret, 15, 4) # 16
      Ops.set(ret, 16, 3) # 17
      Ops.set(ret, 17, 2) # 18
      Ops.set(ret, 18, 1) # 19
      Ops.set(ret, 19, 3) # 20
      Ops.set(ret, 20, 4) # 21
      Ops.set(ret, 21, 2) # 22
      Ops.set(ret, 22, 1) # 23
      Ops.set(ret, 23, 3) # 24
      Ops.set(ret, 24, 1) # 25
      Ops.set(ret, 25, 4) # 26
      Ops.set(ret, 26, 1) # 27
      Ops.set(ret, 27, 4) # 28
      Ops.set(ret, 28, 2) # 29
      Ops.set(ret, 29, 1) # 30
      Ops.set(ret, 30, 3) # 31
      Ops.set(ret, 31, 2) # 32
      Ops.set(ret, 32, 4) # 33
      Ops.set(ret, 33, 2) # 34
      Ops.set(ret, 34, 3) # 35
      Ops.set(ret, 35, 1) # 36
      Ops.set(ret, 36, 2) # 37
      Ops.set(ret, 37, 4) # 38
      Ops.set(ret, 38, 3) # 39
      Ops.set(ret, 39, 1) # 40
      Ops.set(ret, 40, 2) # 41
      Ops.set(ret, 41, 4) # 42
      Ops.set(ret, 42, 3) # 43
      Ops.set(ret, 43, 4) # 44
      Ops.set(ret, 44, 2) # 45
      Ops.set(ret, 45, 1) # 46
      Ops.set(ret, 46, 3) # 47
      Ops.set(ret, 47, 2) # 48
      Ops.set(ret, 48, 1) # 49
      Ops.set(ret, 49, 4) # 50
      Ops.set(ret, 50, 3) # 51
      Ops.set(ret, 51, 4) # 52
      Ops.set(ret, 52, 3) # 53
      Ops.set(ret, 53, 3) # 54
      Ops.set(ret, 54, 1) # 55
      Ops.set(ret, 55, 2) # 56
      Ops.set(ret, 56, 4) # 57
      Ops.set(ret, 57, 3) # 58
      Ops.set(ret, 58, 1) # 59
      Ops.set(ret, 59, 2) # 60
      Ops.set(ret, 60, 1) # 61
      Ops.set(ret, 61, 4) # 62
      Ops.set(ret, 62, 3) # 63
      Ops.set(ret, 63, 4) # 64
      Ops.set(ret, 64, 1) # 65
      Ops.set(ret, 65, 2) # 66
      Ops.set(ret, 66, 4) # 67
      Ops.set(ret, 67, 3) # 68
      Ops.set(ret, 68, 1) # 69
      Ops.set(ret, 69, 4) # 70
      Ops.set(ret, 70, 2) # 71
      Ops.set(ret, 71, 3) # 72
      Ops.set(ret, 72, 1) # 73
      Ops.set(ret, 73, 4) # 74
      Ops.set(ret, 74, 2) # 75
      Ops.set(ret, 75, 1) # 76
      Ops.set(ret, 76, 3) # 77
      Ops.set(ret, 77, 2) # 78
      Ops.set(ret, 78, 4) # 79
      Ops.set(ret, 79, 3) # 80
      Ops.set(ret, 80, 2) # 81
      Ops.set(ret, 81, 1) # 82
      Ops.set(ret, 82, 3) # 83
      Ops.set(ret, 83, 4) # 84
      Ops.set(ret, 84, 3) # 85
      Ops.set(ret, 85, 1) # 86
      Ops.set(ret, 86, 3) # 87
      Ops.set(ret, 87, 4) # 88

      deep_copy(ret)
    end

    # UnObfuscates a string
    def UnObfuscate(str)
      # Initialize obfuscation tables
      tMap = LoadDeTranslationMap()
      bitshift = LoadBitshiftList()

      split = Builtins.splitstring(str, ",")
      sizePword = Builtins.size(split)
      character = ""
      cnt = 0
      val = 0
      clearstring = ""

      Builtins.foreach(split) do |intVal|
        val = Builtins.tointeger(intVal)
        val = Ops.shift_right(val, Ops.get(bitshift, cnt, 0))
        val = Ops.subtract(val, sizePword)
        character = Ops.get_string(tMap, val, "")
        clearstring = Ops.add(clearstring, character)
        cnt = Ops.add(cnt, 1)
      end

      clearstring
    end


    # Obfuscates a string
    def Obfuscate(str)
      # Initialize obfuscation tables
      tMap = LoadTranslationMap()
      bitshift = LoadBitshiftList()

      sizePword = Builtins.size(str)

      workStr = str
      character = ""
      cnt = 0
      val = 0
      obfstring = ""
      while Ops.greater_than(Builtins.size(workStr), 0)
        character = Builtins.substring(workStr, 0, 1)
        val = Ops.get_integer(tMap, character, 0)
        val = Ops.add(val, sizePword)
        val = Ops.shift_left(val, Ops.get(bitshift, cnt, 0))

        if cnt != 0
          obfstring = Ops.add(obfstring, Builtins.sformat(",%1", val))
        else
          obfstring = Ops.add(obfstring, Builtins.sformat("%1", val))
        end

        workStr = Builtins.substring(workStr, 1)
        cnt = Ops.add(cnt, 1)
      end

      obfstring
    end

    # Validates legal password characters
    def ValidatePassword(pw1)
      valid_password_chars = "-0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!@\#$%^&*~(),;:._+/|?{}=[]"
      if pw1 == nil || pw1 == ""
        Report.Error(
          _(
            "The password is missing.\n" +
              "Please try again (e.g. sudo yast2 ./create-answer-file \"password\" \"optional_DSFW_Domain_Admin_password\"\n" +
              "\n" +
              "Use of '!' requires escaping (e.g. \\!)\n" +
              "Use of '-' at beginning of password requires escaping (e.g. \\-)"
          )
        )
        return false
      elsif Builtins.findfirstnotof(pw1, valid_password_chars) != nil
        Report.Error(
          _(
            "The password may only contain the following characters:\n" +
              " 0..9, a..z, A..Z, and any of \"#*,.;:._-+~!$%&/|?{[()]}\".\n" +
              "Use of '!' requires escaping (e.g. \\!)\n" +
              "Use of '-' at beginning of password requires escaping (e.g. \\-)\n" +
              "Please try again."
          )
        )
        return false
      end
      true
    end
    def validate_password(clear_password, dsfw_clear_password)
      # Make sure the password only contains valid characters
      return false if ValidatePassword(clear_password) == false

      if dsfw_clear_password != nil && dsfw_clear_password != ""
        return false if ValidatePassword(dsfw_clear_password) == false
      end

      # Obfuscate the password
      obf_password = Obfuscate(clear_password)
      if dsfw_clear_password != nil && dsfw_clear_password != ""
        dsfw_obf_password = Obfuscate(dsfw_clear_password)
      end

      # Write "answer" file with obfuscated passwords
      if obf_password != nil && obf_password != ""
        writeStr = obf_password

        # Start out with a blank file.
        ret = SCR.Write(path(".target.string"), "answer", "")

        if dsfw_obf_password != nil && dsfw_obf_password != ""
          writeStr = Ops.add(Ops.add(writeStr, "\n----\n"), dsfw_obf_password)
        end
        # Write the password data
        ret = SCR.Write(
          path(".target.string"),
          "answer",
          Ops.add(writeStr, "\n")
        )

        # Flush the cache
        ret = SCR.Write(path(".target.string"), "answer", nil)

        clearPass = UnObfuscate(obf_password) # Get the first password
        msg = Builtins.sformat(
          _("Decrypted password (sanity check) is '%1'. Size is %2"),
          clearPass,
          Builtins.size(clearPass)
        )

        if dsfw_obf_password != nil && dsfw_obf_password != ""
          dsfwClearPass = UnObfuscate(dsfw_obf_password) # Get the dsfw password
          msg = Builtins.sformat(
            _(
              "Decrypted passwords (sanity check) are '%1', size is %2, '%3', size is %4"
            ),
            clearPass,
            Builtins.size(clearPass),
            dsfwClearPass,
            Builtins.size(dsfwClearPass)
          )
        end

        Popup.Message(msg)
        return false
      end

      nil
    end
  end
end

Yast::CreateAnswerFileClient.new.main
