# encoding: utf-8

# -------------------------------------------------------------------------------
# Copyright 2024 Open Text.
#
# The only warranties for products and services of Open Text and its
# affiliates and licensors (“Open Text”) are as may be set forth in the
# express warranty statements accompanying such products and services.
# Nothing herein should be construed as constituting an additional
# warranty. Open Text shall not be liable for technical or editorial
# errors or omissions contained herein. The information contained herein
# is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains
# confidential information and a valid license is required for possession,
# use or copying. If this work is provided to the U.S. Government,
# consistent with FAR 12.211 and 12.212, Commercial Computer Software,
# Computer Software Documentation, and Technical Data for Commercial Items
# are licensed to the U.S. Government under vendor's standard commercial
# license.
# -------------------------------------------------------------------------------
# File:        clients/inst_check_ip.ycp
# Package:     novell-common
# Summary:     Check for static ip and valid host name and dns name
# - warn user of any errors
#
module Yast
  class InstCheckIpClient < Client
    def main

      Yast.import "Service"
      Yast.import "Package"
      Yast.import "Label"
      Yast.import "Stage"
      Yast.import "NovellUtils"
      Yast.import "Report"
      Yast.import "NetworkInterfaces"
      Yast.import "FileUtils"

      textdomain "novell-base"

      @msg = ""
      @hostnameErrorMsg = ""
      @staticIPErrorMsg = ""
      @hostnameError = false
      @staticIPError = false

      @ret = :auto

      @args = Convert.to_map(WFM.Args(0))
      return @ret if Ops.get_boolean(@args, "going_back", false) == true

      if FileUtils.Exists("/var/lib/YaST2/OES11_AUTOUPGRADE")
        Builtins.y2milestone("inst_check_ip: autoyast upgrade mode")
        return @ret
      end

      if Package.Installed("novell-web-base") # FIXME: is this valid???  can we be sure this is the only case we need a static IP??  Why not check based on which service is selected for config in NovellAddonInstall??
        # Check for hostname error
        @host = NovellUtils.GetHostname
        Builtins.y2milestone("inst_check_ip:started host = %1", @host)
        if @host == nil || @host == "" || @host == "linux" ||
            @host == "linux.site" || @host == "localhost"
          @hostnameError = true
          @hostnameErrorMsg = Builtins.sformat(
            _(
              "The host name '%1' is not valid.\nPlease go back and select a valid host name.\n"
            ),
            @host
          )
        end

        @iplist = NovellUtils.GetAllStaticIPAddrs
        if Builtins.size(@iplist) == 0
          @staticIPError = true
          @staticIPErrorMsg = _(
            "This machine does not have an IPv4 static IP address.\n" +
              "Without an IPv4 static IP address eDirectory will not work in an OES environment.\n" +
              "Please go back to the 'Network Configuration' section\n" +
              "to configure an IPv4 static IP address. eDirectory configuration will fail on opting to proceed without an IPv4 static IP address."
          )
        end

        if @hostnameError && @staticIPError
          @msg = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(@msg, Builtins.sformat("%1)", 1)),
                  @hostnameErrorMsg
                ),
                "\n\n"
              ),
              Builtins.sformat("%1)", 2)
            ),
            @staticIPErrorMsg
          )
        elsif @hostnameError
          @msg = @hostnameErrorMsg
        elsif @staticIPError
          @msg = @staticIPErrorMsg
        end
        if Ops.greater_than(Builtins.size(@msg), 0)
          # if (Report::AnyQuestion(_("Invalid Network Configuration"), sformat("%1\n%2\n\n", msg, _("Do you want to change your Network Configuration?")), Label::YesButton(), Label::NoButton(), `focus_yes)) ret = `net;
          # else ret = `skip;
          if Report.AnyQuestion(
              _("Invalid Network Configuration"),
              Builtins.sformat(
                "%1\n\n%2\n\n",
                @msg,
                _("Do you want to change your Network Configuration?")
              ),
              Label.YesButton,
              Label.NoButton,
              :focus_yes
            )
            @ret = :back
          end
          Builtins.y2milestone("Stage::stage = %1", Stage.stage)
          if @ret == :back && Stage.normal
            @ret = Convert.to_symbol(WFM.CallFunction("lan"))
          end
          if @ret == :back && Stage.stage == "continue"
            @ret = Convert.to_symbol(WFM.CallFunction("lan"))
          end
        end
      end
      @ret = :auto if @ret == nil

      Builtins.y2milestone("inst_check_ip returned: %1", @ret)
      @ret

      # EOF
    end
  end
end

Yast::InstCheckIpClient.new.main
