# encoding: utf-8

# -------------------------------------------------------------------------------
# (C) Copyright 2019 - 2022 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates
# and licensors (“Micro Focus”) are as may be set forth in the express warranty
# statements accompanying such products and services. Nothing herein should be
# construed as constituting an additional warranty. Micro Focus shall not be
# liable for technical or editorial errors or omissions contained herein. The
# information contained herein is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains confidential
# information and a valid license is required for possession, use or copying. If
# this work is provided to the U.S. Government, consistent with FAR 12.211 and
# 12.212, Commercial Computer Software, Computer Software Documentation, and
# Technical Data for Commercial Items  are licensed to the U.S. Government under
# vendor's standard commercial license.
# -------------------------------------------------------------------------------

# File:	modules/NovellAddonInstall.ycp
# Package:	novell-common
# Summary:	Novell Addon Install
# Authors:	Michael G. Fritch <mgfritch@novell.com>
#
# $Id: NovellAddonInstall.ycp,v 1.1.2.46 2005/10/04 20:38:19 mgfritch Exp $
#
require "yast"

module Yast
  class NovellAddonInstallClass < Module
    def main

      textdomain "novell-base"

      Yast.import "ProductControl"
      Yast.import "Stage"
      Yast.import "Mode"
      Yast.import "Label"
      Yast.import "Report"
      Yast.import "Popup"
      Yast.import "String"
      Yast.import "NovellPatterns"
      Yast.import "FileUtils"

      @oes_product_display_name = _("Open Enterprise Server")

      @control_file = "" # pathname to the control file (e.g. "/usr/share/YaST/control/oes_install.xml")
      @proposal_name = "oes_product" # current OES proposal name as defined in the control file

      @config_dir_path = "/etc/sysconfig/novell/" # string path to the directory where the service(product) config files reside
      @scr_config_path = path(".etc.sysconfig.novell") # SCR agent path for reading/writing the service(product) config files

      @products_to_configure = {}
      NovellAddonInstall()
    end

    #***************************************************************************************
    # Methods for handling the control file
    #***************************************************************************************


    # Load a custom control file (e.g. control.xml) into ProductControl
    #
    # @param control_file string - full pathname to the product control file (e.g. '/usr/share/YaST2/control/oes_install.xml')
    # @return [Boolean] true on success
    def InitProductControl(new_control_file)
      @control_file = new_control_file
      ProductControl.custom_control_file = @control_file

      if !ProductControl.Init
        # popup error message
        # %1 = absolute pathname to a product control file (i.e. '/usr/share/YaST2/control/vm_install.xml')
        Report.Error(
          Builtins.sformat(
            _("Control file %1 was not found."),
            ProductControl.custom_control_file
          )
        )
        return false
      end

      true
    end


    # Returns the pathname to the current addon control file.
    # Example: GetControlFile() -> "/usr/share/YaST2/control/oes_install.xml"
    #
    # @return [String] pathname to the control file
    def GetControlFile
      @control_file
    end


    # Return the yast modules assoiated with a specified proposal
    #
    # @param [String] proposal_name name of proposal
    # @return [Array<String>] proposal module names
    def GetProposalModules(proposal_name)
      Builtins.y2milestone(
        "NovellAddonProduct::GetProposalModules(%1)",
        proposal_name
      )

      proposal_properties = ProductControl.getProposalProperties(
        Stage.stage,
        Mode.mode,
        proposal_name
      )
      proposal_modules = Ops.get_list(
        proposal_properties,
        "proposal_modules",
        []
      )
      proposal_modules = [] if proposal_modules == nil

      Builtins.y2milestone(
        "NovellAddonProduct::GetProposalModules returned: %1",
        proposal_modules
      )
      deep_copy(proposal_modules)
    end



    #***************************************************************************************
    # Methods for handling the product config files
    #***************************************************************************************



    # Read a setting from a config file using a specified Sysconfig Agent
    #
    # @param [Yast::Path] agent_path SCR Sysconfig agent path
    # @param [String] optname setting to read from the config file
    # @return [String] value of the specified setting in the cfg file
    def ReadSysconfigSetting(agent_path, optname)
      optvalue = Convert.to_string(
        SCR.Read(Ops.add(agent_path, Builtins.topath(optname)))
      )

      optvalue = "" if optvalue == nil # avoid nil value

      Builtins.y2milestone(
        "NovellAddonInstall::ReadSysconfigSetting(%1, %2) returned: %3",
        agent_path,
        optname,
        optvalue
      )
      optvalue
    end


    # Read a setting from a specifed config file
    #
    # @param [String] cfg_file config file name
    # @param [String] optname setting name to read from the cfg file
    # @return [String] value of the specified setting in the cfg file
    def ReadSetting(cfg_file, optname)
      Builtins.y2milestone(
        "NovellAddonInstall::ReadSetting(%1, %2)",
        cfg_file,
        optname
      )
      #	path read_path = scr_config_path + .value + topath(config_dir_path + cfg_file) + topath(optname);
      read_path = Ops.add(
        Ops.add(
          Ops.add(@scr_config_path, path(".value")),
          Builtins.topath(cfg_file)
        ),
        Builtins.topath(optname)
      )
      Builtins.y2debug("read_path=%1", read_path)

      optvalue = Convert.to_string(SCR.Read(read_path))
      optvalue = "" if optvalue == nil # avoid nil value

      Builtins.y2milestone(
        "NovellAddonInstall::ReadSetting returned: %1",
        optvalue
      )
      optvalue
    end


    # Returns the list of network dependencies for a given service
    #
    # @param [String] cfg_file config file of the service
    # @return [Array<String>] network service dependencies
    def GetNetworkServices(cfg_file)
      Builtins.y2milestone(
        "NovellAddonInstall::GetNetworkDependencies(%1)",
        cfg_file
      )
      ret = []

      optvalue = ReadSetting(cfg_file, "NETWORK_SERVICES")
      if optvalue != nil && optvalue != ""
        ret = Builtins.splitstring(optvalue, ", ")
        ret = Builtins.filter(ret) { |f| f != nil && f != "" } # remove empty lines
      end

      Builtins.y2milestone(
        "NovellAddonInstall::GetNetworkDependencies returned: %1",
        ret
      )
      deep_copy(ret)
    end


    # Returns true if a given service has been configured
    #
    # @param [String] cfg_file config file of the service
    # @return [Boolean] true if service is configured
    def IsServiceConfigured(cfg_file)
      Builtins.y2milestone(
        "NovellAddonInstall::IsServiceConfigured(%1)",
        cfg_file
      )
      ret = false

      optvalue = ReadSetting(cfg_file, "SERVICE_CONFIGURED")
      Builtins.y2debug("optvalue=%1", optvalue)
      ret = true if optvalue == "yes"

      Builtins.y2milestone(
        "NovellAddonInstall::IsServiceConfigured returned: %1",
        ret
      )
      ret
    end


    # Returns the name of the pattern assosiated with this service
    #
    # @param [String] cfg_file config file of the service
    # @return [Array<String>] pattern names
    def GetPatternName(cfg_file)
      Builtins.y2milestone("NovellAddonInstall::GetPatternName(%1)", cfg_file)
      ret = []

      optvalue = ReadSetting(cfg_file, "PATTERN_NAME")
      Builtins.y2debug("optvalue=%1", optvalue)

      if optvalue != nil && optvalue != ""
        ret = Builtins.splitstring(optvalue, ", ")
        ret = Builtins.filter(ret) { |f| f != nil && f != "" } # remove empty lines
      end

      Builtins.y2milestone(
        "NovellAddonInstall::GetPatternName returned: %1",
        ret
      )
      deep_copy(ret)
    end


    # Returns a list of names of the YaST clients that must be
    # configured before this this service can be configured
    #
    # @param [String] cfg_file config file of the service
    # @return [Array<String>] YaST client names
    def GetConfigDependencies(cfg_file)
      Builtins.y2milestone(
        "NovellAddonInstall::GetConfigDependencies(%1)",
        cfg_file
      )
      ret = []

      optvalue = ReadSetting(cfg_file, "CONFIG_DEPENDENCIES")
      Builtins.y2debug("optvalue=%1", optvalue)

      optvalue = String.Replace(optvalue, " ", "")
      if optvalue != nil && optvalue != ""
        ret = Builtins.splitstring(optvalue, ",")
        ret = Builtins.filter(ret) { |f| f != nil && f != "" } # remove empty lines
      end

      Builtins.y2milestone(
        "NovellAddonInstall::GetConfigDependencies returned: %1",
        ret
      )
      deep_copy(ret)
    end


    # Returns the name of the YaST client assosiated with this service
    #
    # @param [String] cfg_file config file of the service
    # @return [String] YaST client name
    def GetYaSTClient(cfg_file)
      Builtins.y2milestone("NovellAddonInstall::GetYaSTClient(%1)", cfg_file)
      ret = ""

      optvalue = ReadSetting(cfg_file, "YAST_CLIENT")
      Builtins.y2debug("optvalue=%1", optvalue)
      ret = optvalue if optvalue != nil

      Builtins.y2milestone(
        "NovellAddonInstall::GetYaSTClient returned: %1",
        ret
      )
      ret
    end


    # Reads a specifed config file settings
    #
    # @param [String] cfg_file config file location
    # @return [Hash{String => Object}] of config file settings
    def ReadConfigFileSettings(cfg_file)
      Builtins.y2milestone(
        "NovellAddonInstall::ReadConfigFileSettings(%1)",
        cfg_file
      )
      ret = {}

      # list<string> optnames = (list<string>) SCR::Read(scr_config_path + .all + topath(config_dir_path + cfg_file));
      # 	foreach (string optname, optnames, {
      # 		if (optname != nil && optname !="") {
      # 			ret[optname] = ReadSetting(cfg_file, optname);
      # 		}
      # 	});

      if cfg_file != nil && cfg_file != "" &&
          Ops.greater_or_equal(SCR.Read(path(".target.size"), cfg_file), 0)
        Ops.set(ret, "config_file", cfg_file)
        Ops.set(ret, "NETWORK_SERVICES", GetNetworkServices(cfg_file))
        Ops.set(ret, "SERVICE_CONFIGURED", IsServiceConfigured(cfg_file))
        Ops.set(ret, "PATTERN_NAME", GetPatternName(cfg_file))
        Ops.set(ret, "YAST_CLIENT", GetYaSTClient(cfg_file))
        Ops.set(ret, "CONFIG_DEPENDENCIES", GetConfigDependencies(cfg_file))
      end

      Builtins.y2milestone(
        "NovellAddonInstall::ReadConfigFileSettings returned: %1",
        ret
      )
      deep_copy(ret)
    end


    # Reads the config file directory(config_dir_path) for configurable OES add-on services (products)
    #
    # @return [Array<String>] of OES add-on products
    def ReadAllConfigFileSettings
      Builtins.y2milestone("NovellAddonInstall::ReadAllConfigFileSettings()")
      ret = []

      cfg_file_list = Convert.convert(
        SCR.Read(path(".target.dir"), @config_dir_path),
        :from => "any",
        :to   => "list <string>"
      )
      Builtins.y2milestone("cfg_file_list=%1", cfg_file_list)
      if cfg_file_list != nil &&
          Ops.greater_than(Builtins.size(cfg_file_list), 0)
        Builtins.foreach(cfg_file_list) do |cfg_file|
          first_char = Builtins.substring(cfg_file, 0, 1)
          last_char = Builtins.substring(
            cfg_file,
            Ops.subtract(Builtins.size(cfg_file), 1),
            Builtins.size(cfg_file)
          )
          if first_char == "." || last_char == "~"
            # ignore all files that start with '.' or end with '~'
            next
          else
            new_settings = ReadConfigFileSettings(
              Ops.add(@config_dir_path, cfg_file)
            )
            if new_settings != nil &&
                Ops.greater_than(Builtins.size(new_settings), 0)
              ret = Builtins.add(ret, new_settings)
            end
          end
        end
      end

      Builtins.y2milestone(
        "NovellAddonInstall::ReadAllConfigFileSettings() returned: %1",
        ret
      )
      deep_copy(ret)
    end

    def StripLeading(str, stripChars)
      while Ops.greater_than(Builtins.size(str), 1)
        check = Builtins.substring(str, 0, 1)
        if Builtins.findfirstof(stripChars, check) != nil
          str = Builtins.substring(str, 1)
        else
          break
        end
      end
      str
    end

    def StripTrailing(str, stripChars)
      while Ops.greater_than(Builtins.size(str), 1)
        check = Builtins.substring(str, Ops.subtract(Builtins.size(str), 1), 1)
        if Builtins.findfirstof(stripChars, check) != nil
          str = Builtins.substring(str, 0, Ops.subtract(Builtins.size(str), 1))
        else
          break
        end
      end
      str
    end

    def Strip(str, stripChars)
      str = StripLeading(str, stripChars)
      str = StripTrailing(str, stripChars)
      str
    end

    # Handle the case where there is CR/LF at end of value which pushes the close quote to the next line
    def FixLine(line)
      workstr = Strip(line, " ")
      check = Builtins.substring(workstr, 0, 1)
      if Builtins.findfirstof("#", check) != nil # we don't care about comment lines
        return line
      end

      parts = Builtins.splitstring(workstr, "=")
      if Ops.greater_or_equal(Builtins.size(parts), 2)
        # CONFIG_VALUE = "test"
        # CONFIG-VALUE = "test     // we are looking for this second case with the missing quote
        check = Builtins.substring(
          workstr,
          Ops.subtract(Builtins.size(workstr), 1),
          1
        )
        if Builtins.findfirstof("\"", check) == nil
          workstr = Ops.add(workstr, "\"") # add the missing close quote
        end
      else
        workstr = line
      end

      workstr
    end

    # Read settings from a config file
    #
    # @param [String] cfg_file config file location
    # @return [Hash{String => String}] of config file settings
    def GetSysconfigMap(cfg_file)
      ret = {}
      Builtins.y2debug("creating settings map for %1", cfg_file)
      if cfg_file != nil && cfg_file != "" &&
          Ops.greater_or_equal(FileUtils.GetSize(cfg_file), 0) # ensure the config file exists!
        pattern = "^([_[:alnum:]-]+)[[:space:]]*=[[:space:]]*\"(.*)\"[[:space:]]*#?.*$"
        contents = Convert.to_string(SCR.Read(path(".target.string"), cfg_file))

        lines = Builtins.splitstring(contents, "\n")
        Builtins.foreach(lines) do |line|
          if Builtins.size(line) > 0
            Builtins.y2debug("checking line: %1", line)
            line = FixLine(line) # Handle case of extra CR at end of line
            parts = Builtins.regexptokenize(line, pattern)
            Builtins.y2debug("tokenized parts: %1", parts)
            if Builtins.size(parts) == 2
              Ops.set(
                ret,
                Ops.get(parts, 0, ""),
                Builtins.deletechars(Ops.get(parts, 1, ""), "\"'")
              )
            end
	  end
        end
      end
      Builtins.y2milestone("sysconfig values: %1", ret)
      deep_copy(ret)
    end

    # Read settings from a config file
    #
    # @param [String] cfg_file config file location
    # @return [Hash{String => String}] of config file settings
    def GetFileContentMap(cfg_file)
      ret = {}
      Builtins.y2debug("creating settings map for %1", cfg_file)
      if cfg_file != nil && cfg_file != "" &&
          Ops.greater_or_equal(FileUtils.GetSize(cfg_file), 0) # ensure the config file exists!
        pattern = "^([_[:alnum:]-]+)[[:space:]]*=[[:space:]]*(.*)[[:space:]]*#?.*$"
        contents = Convert.to_string(SCR.Read(path(".target.string"), cfg_file))

        lines = Builtins.splitstring(contents, "\n")
        Builtins.foreach(lines) do |line|
          if Builtins.size(line) > 0
            Builtins.y2debug("checking line: %1", line)
            line = FixLine(line) # Handle case of extra CR at end of line
            parts = Builtins.regexptokenize(line, pattern)
            Builtins.y2debug("tokenized parts: %1", parts)
            if Builtins.size(parts) == 2
              Ops.set(
                ret,
                Ops.get(parts, 0, ""),
                Builtins.deletechars(Ops.get(parts, 1, ""), "\"'")
              )
            end
	  end
        end
      end
      Builtins.y2milestone("sysconfig values: %1", ret)
      deep_copy(ret)
    end

    #***************************************************************************************
    # Methods for handling the product selection (installation/configuration)
    #***************************************************************************************


    # Retrieve the name(rich_text_title) of a YaST module in a
    # user readable form for display on the UI.
    #
    # @param [String] module_name name of the YaST module
    # @return [String] module name for display on the UI
    def GetModuleDisplayName(module_name)
      Builtins.y2milestone("NovellAddonInstall::GetModuleName(%1)", module_name)

      ret = ""
      proposal_descr = Convert.to_map(
        WFM.CallFunction(
          Ops.add(module_name, "_proposal"),
          ["Description", { "force_proposal" => true }]
        )
      ) # call the YaST module proposal client
      ret = Ops.get_string(proposal_descr, "rich_text_title", "")

      Builtins.y2milestone(
        "NovellAddonInstall::GetModuleName returned: %1",
        ret
      )
      ret
    end


    # Return the information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of information about the product
    def GetProductInfo(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductInfo(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_info"), ["Info", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductInfo returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product

    def GetProductUpgradeInfo(upgrade_from, module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfo(%1)",
        module_name
      )

      if upgrade_from == "OES2015SP1"
        ret = GetProductUpgradeInfoOES2015SP1(module_name)
      elsif upgrade_from == "OES2018"
        ret = GetProductUpgradeInfoOES2018(module_name)
      elsif upgrade_from == "OES2018SP1"
        ret = GetProductUpgradeInfoOES2018SP1(module_name)
      elsif upgrade_from == "OES2018SP2"
        ret = GetProductUpgradeInfoOES2018SP2(module_name)
      elsif upgrade_from == "OES2018SP3"
        ret = GetProductUpgradeInfoOES2018SP3(module_name)
      end

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfo for %1 upgrade_from %2 returned: %3",
        module_name,
        upgrade_from,
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product

    def GetProductUpgradeInfoOES2015SP1(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2015SP1(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_upgrade"), ["Info_oes2015_sp1", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2015SP1 returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product

    def GetProductUpgradeInfoOES2018(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_upgrade"), ["Info_oes2018", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018 returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product

    def GetProductUpgradeInfoOES2018SP1(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018SP1(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_upgrade"), ["Info_oes2018_sp1", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018SP1 returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product

    def GetProductUpgradeInfoOES2018SP2(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018SP2(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_upgrade"), ["Info_oes2018_sp2", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018SP2 returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product

    def GetProductUpgradeInfoOES2018SP3(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018SP3(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_upgrade"), ["Info_oes2018_sp3", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES2018SP3 returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Return the upgrade information about a product from it's config file
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] of upgrade information about the product
    # For OES2023 and onwards upgrade

    def GetProductUpgradeInfoOES(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES(%1)",
        module_name
      )

      ret = Convert.to_map(
        WFM.CallFunction(Ops.add(module_name, "_upgrade"), ["Info_oes", {}])
      )

      Builtins.y2milestone(
        "NovellAddonInstall::GetProductUpgradeInfoOES returned: %1",
        ret
      )
      deep_copy(ret)
    end

    # Check if a given product is selected for configuration
    #
    # @param [String] product_name name of the product
    # @return [Boolean] true if the product is selected
    def IsProductSelected(product_name)
      Builtins.y2milestone(
        "NovellAddonInstall::IsProductSelected(%1)",
        product_name
      )

      ret = true
      info = GetProductInfo(product_name)
      patterns_list = Ops.get_list(info, "PATTERN_NAME", [])
      Builtins.foreach(patterns_list) do |pattern|
        ret = ret && NovellPatterns.IsPatternInstalled(pattern)
      end

      Builtins.y2milestone(
        "NovellAddonInstall::IsProductSelected returned: %1",
        ret
      )
      ret
    end


    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    def IsProductUpgrade(upgrade_from, module_name)
      Builtins.y2milestone("IsProductUpgrade(%1)", module_name)
      ret = false
      if upgrade_from == "OES2015SP1"
        ret = IsProductUpgradeOES2015SP1(module_name)
      elsif upgrade_from == "OES2018"
        ret = IsProductUpgradeOES2018(module_name)
      elsif upgrade_from == "OES2018SP1"
        ret = IsProductUpgradeOES2018SP1(module_name)
      elsif upgrade_from == "OES2018SP2"
        ret = IsProductUpgradeOES2018SP2(module_name)
      elsif upgrade_from == "OES2018SP3"
        ret = IsProductUpgradeOES2018SP3(module_name)
      end
      Builtins.y2milestone("IsProductUpgrade for %1 upgrade from %2 returned: %3", module_name, upgrade_from, ret)
      ret
    end

    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    def IsProductUpgradeOES2015SP1(module_name)
      Builtins.y2milestone("IsProductUpgradeOES2015SP1(%1)", module_name)
      ret = false

      info = GetProductInfo(module_name) # Get current OES 2015 SP1 roduct info.
      upgrade_info = GetProductUpgradeInfoOES2015SP1(module_name) # Get old OES2015SP1 product info.

      product_configured = "no"
      if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
        product_configured = "yes"
      end # Current product is already configured - don't do upgrade.

      old_product_configured = "no"
      if Ops.get_boolean(upgrade_info, "SERVICE_CONFIGURED", false) == true
        old_product_configured = "yes"
      end # Old OES2015 product is configured - we can do an upgrade.

      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      Builtins.y2milestone(
        "IsProductUpgradeOES2015SP1: old file - %1",
        old_sysconfig_file
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2015SP1: product configured - %1",
        product_configured
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2015SP1: old_product configured - %1",
        old_product_configured
      )

      if Ops.greater_than(FileUtils.GetSize(old_sysconfig_file), 0)
        if product_configured == "no" && old_product_configured == "yes"
          ret = true
        end
      end
      Builtins.y2milestone("IsProductUpgradeOES2015SP1 returned: %1", ret)
      ret
    end

    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    def IsProductUpgradeOES2018(module_name)
      Builtins.y2milestone("IsProductUpgradeOES2018(%1)", module_name)
      ret = false

      info = GetProductInfo(module_name) # Get current OES 2018 roduct info.
      upgrade_info = GetProductUpgradeInfoOES2018(module_name) # Get old OES2018 product info.

      product_configured = "no"
      if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
        product_configured = "yes"
      end # Current product is already configured - don't do upgrade.

      old_product_configured = "no"
      if Ops.get_boolean(upgrade_info, "SERVICE_CONFIGURED", false) == true
        old_product_configured = "yes"
      end # Old OES2018 product is configured - we can do an upgrade.

      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      Builtins.y2milestone(
        "IsProductUpgradeOES2018: old file - %1",
        old_sysconfig_file
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018: product configured - %1",
        product_configured
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018: old_product configured - %1",
        old_product_configured
      )

      if Ops.greater_than(FileUtils.GetSize(old_sysconfig_file), 0)
        if product_configured == "no" && old_product_configured == "yes"
          ret = true
        end
      end
      Builtins.y2milestone("IsProductUpgradeOES2018 returned: %1", ret)
      ret
    end

    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    def IsProductUpgradeOES2018SP1(module_name)
      Builtins.y2milestone("IsProductUpgradeOES2018SP1(%1)", module_name)
      ret = false

      info = GetProductInfo(module_name) # Get current OES 2018 SP2 product info.
      upgrade_info = GetProductUpgradeInfoOES2018SP1(module_name) # Get old OES2018SP1 product info.

      product_configured = "no"
      if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
        product_configured = "yes"
      end # Current product is already configured - don't do upgrade.

      old_product_configured = "no"
      if Ops.get_boolean(upgrade_info, "SERVICE_CONFIGURED", false) == true
        old_product_configured = "yes"
      end # Old OES2018SP1 product is configured - we can do an upgrade.

      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP1: old file - %1",
        old_sysconfig_file
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP1: product configured - %1",
        product_configured
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP1: old_product configured - %1",
        old_product_configured
      )

      if Ops.greater_than(FileUtils.GetSize(old_sysconfig_file), 0)
        if product_configured == "no" && old_product_configured == "yes"
          ret = true
        end
      end
      Builtins.y2milestone("IsProductUpgradeOES2018SP1 returned: %1", ret)
      ret
    end

    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    def IsProductUpgradeOES2018SP2(module_name)
      Builtins.y2milestone("IsProductUpgradeOES2018SP2(%1)", module_name)
      ret = false

      info = GetProductInfo(module_name) # Get current OES 2018 SP3 product info.
      upgrade_info = GetProductUpgradeInfoOES2018SP2(module_name) # Get old OES2018SP2 product info.

      product_configured = "no"
      if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
        product_configured = "yes"
      end # Current product is already configured - don't do upgrade.

      old_product_configured = "no"
      if Ops.get_boolean(upgrade_info, "SERVICE_CONFIGURED", false) == true
        old_product_configured = "yes"
      end # Old OES2018SP2 product is configured - we can do an upgrade.

      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP2: old file - %1",
        old_sysconfig_file
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP2: product configured - %1",
        product_configured
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP2: old_product configured - %1",
        old_product_configured
      )

      if Ops.greater_than(FileUtils.GetSize(old_sysconfig_file), 0)
        if product_configured == "no" && old_product_configured == "yes"
          ret = true
        end
      end
      Builtins.y2milestone("IsProductUpgradeOES2018SP2 returned: %1", ret)
      ret
    end

    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    def IsProductUpgradeOES2018SP3(module_name)
      Builtins.y2milestone("IsProductUpgradeOES2018SP3(%1)", module_name)
      ret = false

      info = GetProductInfo(module_name) # Get current OES 2018 SP3 product info.
      upgrade_info = GetProductUpgradeInfoOES2018SP3(module_name) # Get old OES2018SP3 product info.

      product_configured = "no"
      if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
        product_configured = "yes"
      end # Current product is already configured - don't do upgrade.

      old_product_configured = "no"
      if Ops.get_boolean(upgrade_info, "SERVICE_CONFIGURED", false) == true
        old_product_configured = "yes"
      end # Old OES2018SP3 product is configured - we can do an upgrade.

      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP3: old file - %1",
        old_sysconfig_file
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP3: product configured - %1",
        product_configured
      )
      Builtins.y2milestone(
        "IsProductUpgradeOES2018SP3: old_product configured - %1",
        old_product_configured
      )

      if Ops.greater_than(FileUtils.GetSize(old_sysconfig_file), 0)
        if product_configured == "no" && old_product_configured == "yes"
          ret = true
        end
      end
      Builtins.y2milestone("IsProductUpgradeOES2018SP3 returned: %1", ret)
      ret
    end

    # Determine if this product should be upgraded
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true if a previous version of the module
    # (i.e. the old sysconfig file) exists, AND the new version
    # has not yet been configured
    # For OES2023 and onwards upgrade
    def IsProductUpgradeOES(module_name)
      Builtins.y2milestone("IsProductUpgradeOES(%1)", module_name)
      ret = false

      info = GetProductInfo(module_name)
      upgrade_info = GetProductUpgradeInfoOES(module_name)

      old_product_configured = "no"
      if Ops.get_boolean(upgrade_info, "SERVICE_CONFIGURED", false) == true
        old_product_configured = "yes"
      end

      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      Builtins.y2milestone(
        "IsProductUpgrade: old file - %1",
        old_sysconfig_file
      )
      Builtins.y2milestone(
        "IsProductUpgrade: old_product configured - %1",
        old_product_configured
      )

      if Ops.greater_than(FileUtils.GetSize(old_sysconfig_file), 0)
        if old_product_configured == "yes"
          ret = true
        end
      end
      Builtins.y2milestone("IsProductUpgradeOES returned: %1", ret)
      ret
    end


    # Translate settings map for product upgrade
    #
    # @param [String] module_name name of the module
    # @return [Boolean] true the translation was successful
    def TranslateSettings(module_name)
      Builtins.y2milestone("TranslateSettings(%1)", module_name)
      ret = false
      upgrade_info = GetProductUpgradeInfoOES(module_name)
      old_sysconfig_file = Ops.get_string(upgrade_info, "config_file", "")
      old_settings = GetSysconfigMap(old_sysconfig_file)
      results = Convert.to_map(
        WFM.CallFunction(
          Ops.add(module_name, "_upgrade"),
          ["Translate", old_settings]
        )
      )
      ret = Ops.get_boolean(results, "success", true)
      Builtins.y2milestone("TranslateSettings returned: %1", ret)
      ret
    end


    # Check/Resolve all dependencies for a given YaST module (product)
    #
    # 1. Ensure that the desired product is not yet configured or is selected for non-destructive re-install (SERVICE_CONFIGURED)
    # 2. Ensure all required patterns are installed (PATTERN_NAME)
    # 3. Ensure all required network services are accessable (NETWORK_SERVICES)
    #
    # @param [String] module_name string name of the YaST module
    # @return [Boolean] true is all dependencies are resolved/correct
    def SolveProductDependencies(module_name)
      Builtins.y2milestone(
        "NovellAddonInstall::SolveProductDependencies(%1)",
        module_name
      )

      ret = true
      info = GetProductInfo(module_name) # Get the product info

      # 2. Ensure all required patterns are installed (PATTERN_NAME)
      if ret && Mode.normal
        software_patterns = Ops.get_list(info, "PATTERN_NAME", [])
        if NovellPatterns.EnsurePatternsInstalled(software_patterns) == false
          return false
        end
      end

      # 3. Ensure all required network services are installed (NETWORK_SERVICES)
      if ret
        network_services = Ops.get_list(info, "NETWORK_SERVICES", [])

        # remove items from the network service dependencies list that are already configured or selected for configuration
        Builtins.foreach(network_services) do |service|
          if IsProductSelected(service)
            # service is selected for configuration
            network_services = Builtins.filter(network_services) do |s|
              s != service
            end # remove it from the list of required network services
          end
          service_info = GetProductInfo(service)
          if Ops.get_boolean(service_info, "SERVICE_CONFIGURED", false)
            # service is already configured locally
            network_services = Builtins.filter(network_services) do |s|
              s != service
            end # remove it from the list of required network services
          end
        end

        # display popup of unresolved required network services
        if network_services != nil && network_services != []
          # Generate a list of user readable product names to display in the Popup.
          display_names_list = []
          Builtins.foreach(network_services) do |service|
            display_names_list = Builtins.add(
              display_names_list,
              GetModuleDisplayName(service)
            )
          end
          display_names_list = Builtins.filter(display_names_list) do |item|
            item != nil && item != ""
          end # Remove all empty list items

          # Display a Continue/Cancel popup
          if Ops.greater_than(Builtins.size(display_names_list), 0)
            ret = ret &&
              Popup.AnyQuestionRichText(
                _("Unresolved Network Dependencies"),
                # rich text message
                # %1 product neme with the unresolved dependencies
                # %2 list of product names
                Builtins.sformat(
                  _(
                    "<p><b>%1</b> can not be configured unless the following products are installed somewhere in your network:</p><p><b>%2</b></p><p>Are they installed in your network?</p>"
                  ),
                  GetModuleDisplayName(module_name),
                  Builtins.mergestring(display_names_list, "<br>")
                ),
                50,
                15,
                # popup message button
                Label.ContinueButton,
                Label.CancelButton,
                :no
              )
          end
        end
      end


      Builtins.y2milestone(
        "NovellAddonInstall::SolveProductDependencies(%1) returned: %2",
        module_name,
        ret
      )
      ret
    end


    # Return a list of modules that must be configured before
    # the specified module may be configured, that are currently
    # not set to be configured.
    #
    # @param [String] module_name name of product's yast module
    # @return [Array<String>] unsatisfied configuration dependencies
    def GetUnsatisfiedConfigDependencies(module_name)
      ret = []
      deps = Ops.get_list(
        GetProductInfo(module_name),
        "CONFIG_DEPENDENCIES",
        []
      )
      Builtins.y2milestone("%1 dependencies: %2", module_name, deps)
      ret = Builtins.filter(deps) do |dep|
        !(Ops.get_boolean(GetProductInfo(dep), "SERVICE_CONFIGURED", false) ||
          Ops.get(@products_to_configure, dep, true))
      end
      Builtins.y2milestone(
        "unsatisfied %1 config dependencies: %2",
        module_name,
        ret
      )
      deep_copy(ret)
    end


    # Return an HTML listing of unsatisfied configuration depencencies
    #
    # @param [String] module_name name of product's yast module
    # @return [String] unsatisfied configuration dependencies in HTML
    def GetUnsatisfiedConfigHTML(module_name)
      ret = ""
      unsatisfied = GetUnsatisfiedConfigDependencies(module_name)
      if unsatisfied != nil && unsatisfied != []
        ret = _("The following components must be configured:")
        ret = Ops.add(ret, "<ul>")
        Builtins.foreach(unsatisfied) do |dep|
          ret = Ops.add(
            Ops.add(Ops.add(ret, "\n<li>"), GetModuleDisplayName(dep)),
            "</li>"
          )
        end 

        ret = Ops.add(ret, "\n</ul>")
      end
      Builtins.y2milestone(
        "unsatisfied %1 config dependencies (HTML): %2",
        module_name,
        ret
      )
      ret
    end


    # Construct a map to be used by <module>_proposal
    #
    # @param [String] module_name name of product's yast module
    # @return [Hash] containing all proposal values, i.e.
    #  "preformatted_proposal" : <string>,
    #  "warning_level" : <symbol>,
    #  "warning" : <string>,
    #  "links" : <list>
    def ConstructProposal(module_name, current_proposal)
      ret = {}
      proposal = current_proposal
      warning = ""
      warning_level = nil
      now_link = Ops.add(module_name, "--configure_now")
      later_link = Ops.add(module_name, "--configure_later")
      html = "%1<a href=\"%2\">%3</a>%4</p>"

      # if upgrade is from OES2023 all sysconfig will be same and SERVICE_CONFIGURED will be true
      # and if it is true and Mode is update or autopgrade then go with first if
      if Ops.get_boolean(
          GetProductInfo(module_name),
          "SERVICE_CONFIGURED",
          false
        ) && ( Mode.update || Mode.autoupgrade )

        if Ops.get(@products_to_configure, module_name, true)
          proposal = Ops.add(
            Ops.add(
              Builtins.sformat(
                html,
                _("Configure is "),
                later_link,
                _("enabled"),
                ""
              ),
              proposal
            ),
            Ops.greater_than(Builtins.size(proposal), 0) ?
              "----------------------------------------------------------------------</p>" :
              "<br><br>----------------------------------------------------------------------</p>"
          )
        else
          proposal = Ops.add(
            Builtins.sformat(
              html,
              _("Configure is "),
              now_link,
              _("disabled"),
              ""
            ),
            "<br><br>----------------------------------------------------------------------</p>"
          )
        end
      # check to see if the product is installed -- assume it is not
      elsif Ops.get_boolean(
          GetProductInfo(module_name),
          "SERVICE_CONFIGURED",
          false
        )
        # if the product is installed, check to see if it is set for configuration --
        # assume the user will keep existing settings
        if Ops.get(@products_to_configure, module_name, false)
          proposal = Ops.add(
            Ops.add(
              Builtins.sformat(
                html,
                _("Reconfigure is "),
                later_link,
                _("enabled"),
                ""
              ),
              proposal
            ),
            Ops.greater_than(Builtins.size(proposal), 0) ?
              "----------------------------------------------------------------------</p>" :
              "<br><br>----------------------------------------------------------------------</p>"
          )

          Ops.set(@products_to_configure, module_name, true) # be more explict with values in products_to_configure map, don't just assume (#230038)
        else
          proposal = Ops.add(
            Builtins.sformat(
              html,
              _("Reconfigure is "),
              now_link,
              _("disabled"),
              ""
            ),
            "<br><br>----------------------------------------------------------------------</p>"
          )
          Ops.set(@products_to_configure, module_name, false) # be more explict with values in products_to_configure map, don't just assume (#230038)
        end
      else
        if Ops.get(@products_to_configure, module_name, true)
          proposal = Ops.add(
            Ops.add(
              Builtins.sformat(
                html,
                _("Configure is "),
                later_link,
                _("enabled"),
                ""
              ),
              proposal
            ),
            Ops.greater_than(Builtins.size(proposal), 0) ?
              "----------------------------------------------------------------------</p>" :
              "<br><br>----------------------------------------------------------------------</p>"
          )
        else
          proposal = Ops.add(
            Builtins.sformat(
              html,
              _("Configure is "),
              now_link,
              _("disabled"),
              ""
            ),
            "<br><br>----------------------------------------------------------------------</p>"
          )
        end
      end

      # if necessary, validate the proposal parameters
      if !Ops.get_boolean(
          GetProductInfo(module_name),
          "SERVICE_CONFIGURED",
          false
        ) &&
          Ops.get(@products_to_configure, module_name, true) ||
          Ops.get_boolean(
            GetProductInfo(module_name),
            "SERVICE_CONFIGURED",
            false
          ) &&
            Ops.get(@products_to_configure, module_name, false) ||
          (Ops.get_boolean(
            GetProductInfo(module_name),
            "SERVICE_CONFIGURED",
            false
          ) && ( Mode.update || Mode.autoupgrade ))

        # first check for unsatisfied config dependencies
        unsatis = GetUnsatisfiedConfigHTML(module_name)
        if unsatis != nil && unsatis != ""
          warning = unsatis
          warning_level = :blocker
        end

        # now check to make sure the values are valid
        valid_proposal = Convert.to_map(
          WFM.CallFunction(
            Ops.add(module_name, "_proposal"),
            ["ValidateValues", {}]
          )
        )
        if Ops.get_boolean(valid_proposal, "errors", false)
          display_name = GetModuleDisplayName(module_name)
          Builtins.y2error("%1 needs to be configured first", display_name)
          warning = Ops.add(
            Ops.add(warning, display_name),
            _(
              " requires additional configuration information before continuing or disable the configuration."
            )
          )
          warning_level = :blocker
        end
      end

      ret = {
        "preformatted_proposal" => proposal,
        "warning_level"         => warning_level,
        "warning"               => warning,
        "links"                 => [later_link, now_link]
      }

      deep_copy(ret)
    end


    # Constructor - Ensure this modules SCR agents are properly registered.
    def NovellAddonInstall
      # Use the display name of the OES product from the OES inst source data in the zypp database...
      @oes_product_display_name = NovellPatterns.GetProductDisplayName(
        NovellPatterns.oes_addon_name
      ) 

      # // FIXME: The below code is a bad hack!!!  Must talk with SUSE YaST team to solve the problem...
      # // make sure this modules methods for handling the product config files will work...
      # string agent_cfg = "etc_sysconfig_novell.scr";
      # path read_path = .value + topath("/etc/sysconfig/novell/edirectory") + topath("SERVICE_CONFIGURED"); // Test if we can read the eDirectory cfg file
      # NovellUtils::EnsureAgentMounted(agent_cfg, scr_config_path, read_path); // FIXME: this agent is not a sysconfig agent!!!

      nil
    end

    publish :variable => :oes_product_display_name, :type => "string"
    publish :variable => :proposal_name, :type => "string"
    publish :variable => :config_dir_path, :type => "string"
    publish :variable => :scr_config_path, :type => "path"
    publish :variable => :products_to_configure, :type => "map <string, boolean>"
    publish :function => :InitProductControl, :type => "boolean (string)"
    publish :function => :GetControlFile, :type => "string ()"
    publish :function => :GetProposalModules, :type => "list <string> (string)"
    publish :function => :GetNetworkServices, :type => "list <string> (string)"
    publish :function => :IsServiceConfigured, :type => "boolean (string)"
    publish :function => :GetPatternName, :type => "list <string> (string)"
    publish :function => :GetConfigDependencies, :type => "list <string> (string)"
    publish :function => :GetYaSTClient, :type => "string (string)"
    publish :function => :ReadConfigFileSettings, :type => "map <string, any> (string)"
    publish :function => :ReadAllConfigFileSettings, :type => "list <map> ()"
    publish :function => :GetSysconfigMap, :type => "map (string)"
    publish :function => :GetModuleDisplayName, :type => "string (string)"
    publish :function => :GetProductInfo, :type => "map (string)"
    publish :function => :GetProductUpgradeInfo, :type => "map (string, string)"
    publish :function => :GetProductUpgradeInfoOES2015SP1, :type => "map (string)"
    publish :function => :GetProductUpgradeInfoOES2018, :type => "map (string)"
    publish :function => :GetProductUpgradeInfoOES2018SP1, :type => "map (string)"
    publish :function => :GetProductUpgradeInfoOES2018SP2, :type => "map (string)"
    publish :function => :GetProductUpgradeInfoOES2018SP3, :type => "map (string)"
    publish :function => :GetProductUpgradeInfoOES, :type => "map (string)"
    publish :function => :IsProductSelected, :type => "boolean (string)"
    publish :function => :IsProductUpgrade, :type => "boolean (string, string)"
    publish :function => :IsProductUpgradeOES2015SP1, :type => "boolean (string)"
    publish :function => :IsProductUpgradeOES2018, :type => "boolean (string)"
    publish :function => :IsProductUpgradeOES2018SP1, :type => "boolean (string)"
    publish :function => :IsProductUpgradeOES2018SP2, :type => "boolean (string)"
    publish :function => :IsProductUpgradeOES2018SP3, :type => "boolean (string)"
    publish :function => :IsProductUpgradeOES, :type => "boolean (string)"
    publish :function => :TranslateSettings, :type => "boolean (string)"
    publish :function => :SolveProductDependencies, :type => "boolean (string)"
    publish :function => :GetUnsatisfiedConfigDependencies, :type => "list <string> (string)"
    publish :function => :GetUnsatisfiedConfigHTML, :type => "string (string)"
    publish :function => :ConstructProposal, :type => "map (string, string)"
    publish :function => :NovellAddonInstall, :type => "void ()"
  end

  NovellAddonInstall = NovellAddonInstallClass.new
  NovellAddonInstall.main
end
