# encoding: utf-8

# Copyright (c) 2020 Micro Focus or one of its affiliates.
# File:	modules/NovellPatterns.ycp
# Package:	yast2-novell-common
# Summary:	API calls for handling zypp patterns/products
# Authors:	Michael G. Fritch <mgfritch@novell.com>
#
# Used for handling of zypp packages, patterns, and products used by Novell OES components
require "yast"

module Yast
  class NovellPatternsClass < Module
    def main
      Yast.import "UI"
      Yast.import "Pkg"

      textdomain "novell-base"

      Yast.import "Wizard"
      Yast.import "Popup"
      Yast.import "Report"
      Yast.import "Label"
      Yast.import "PackageSystem"
      Yast.import "SlideShow"
      Yast.import "Mode"
      Yast.import "Installation"
      Yast.import "Kernel"
      Yast.import "Directory"
      Yast.import "GetInstArgs"
      Yast.import "Stage"
      Yast.import "ProductControl"
      Yast.import "PackageSlideShow"
      Yast.import "FileUtils"
      #  ***************************************************************
      #  * Output of Pkg::PatternData(pattern_name)
      #  ***************************************************************
      #
      # $[
      # "arch":"i586",
      # "category":"OES Services",
      # "default":false,
      # "description":"FTP services integrated with NetIQ eDirectory.\n",
      # "icon":"",
      # "order":"120",
      # "script":"",
      # "srcid":2,
      # "summary":"Novell FTP",
      # "version":"10-31",
      # "visible":false
      # ]


      #  ***************************************************************
      #  * Output of Pkg::ResolvableDependencies(pattern_name)
      #  ***************************************************************
      #
      # [
      # 	$[
      # 		"arch":"i586",
      # 		"category":"OES Services",
      # 		"default":false,
      # 		"dependencies":[
      # 			$[
      # 				"dep_kind":"provides",
      # 				"name":"novell-sms == 10-15",
      # 				"res_kind":"pattern"
      # 			],
      # 			$[
      # 				"dep_kind":"requires",
      # 				"name":"novell-sms-openwbem-provider",
      # 				"res_kind":"package"
      # 			],
      # 			$[
      # 				"dep_kind":"requires",
      # 				"name":"novell-sms",
      # 				"res_kind":"package"
      # 			],
      # 			$[
      # 				"dep_kind":"requires",
      # 				"name":"novell-base",
      # 				"res_kind":"pattern"
      # 			]
      # 		],
      # 		"description":"Novell Storage Management Services.\n",
      # 		"icon":"",
      # 		"name":"novell-sms",
      # 		"script":"",
      # 		"source":2,
      # 		"status":`available,
      # 		"summary":"Novell SMS",
      # 		"user_visible":true,
      # 		"version":"10-15"
      # 	]
      # ]


      #  ***************************************************************
      #  * Valid dependency kinds
      #  ***************************************************************
      #  *dependencies*
      # 	dep_kind:	provides,
      # 			prerequires,
      # 			requires,
      # 			conflicts,
      # 			obsoletes,
      # 			recommends,
      # 			suggests,
      # 			freshens,
      # 			enhances,
      # 			supplements
      # 	res_kind:	package,
      # 			pattern,
      # 			product,
      # 			patch,
      # 			selection (NOTE: this res_kind will be removed in the future)



      # Name of the OES product inst source as it appears in the zypp database
      @oes_addon_name = "Open_Enterprise_Server"
      @dsfw_install_abort = false
      NovellPatterns()
    end

    #***********************************************************
    # Product query routines
    #***********************************************************


    # Query zypp for the list of all known products (`installed or `available)
    #
    # @return [Array<String>] all product names known by zypp
    def GetAllKnownProducts
      # Query zypp for list of all known products
      product_properties = Pkg.ResolvableDependencies("", :product, "")
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(\"\", `product, \"\") returned: %1",
        product_properties
      )

      # Create the list of known products
      known_products = []
      Builtins.foreach(product_properties) do |product|
        Builtins.y2milestone(
          "NovellPatterns::GetAllKnownProducts(): name %1, status %2",
          Ops.get_string(product, "name", ""),
          Ops.get(product, "status")
        )
        if Ops.get(product, "status") == :available
          known_products = Builtins.add(
            known_products,
            Ops.get_string(product, "name", "")
          )
        end
      end
      known_products = Builtins.toset(known_products)

      Builtins.y2milestone(
        "NovellPatterns::GetAllKnownProducts() returned: %1",
        known_products
      )
      deep_copy(known_products)
    end

    # Given a partial product name, retrive the full product name
    # For Example: "Novell-Open-Enterprise-Server" -> "Novell-Open-Enterprise-Server-i386"
    #
    # @param [String] partial_name regexp partial name of product
    # @return [String] full name of product
    def GetFullProductName(partial_name)
      ret = ""
      all_products = GetAllKnownProducts()
      exp = Builtins.sformat("%1", partial_name) # Partially match the first part of the product name. For example: ".*Open.*Enterprise.*Server.*" -> "Open_Enterprise_Server"

      expUP = Builtins.toupper(exp)
      expLO = Builtins.tolower(exp)
      Builtins.foreach(all_products) do |product_str|
        if Builtins.regexpmatch(product_str, ".*migration.*")
          Builtins.y2warning(
            "WARNING: Found migration product.  Skipping.(%1)",
            product_str
          )
        else
          if Builtins.regexpmatch(product_str, exp) ||
              Builtins.regexpmatch(product_str, expUP) ||
              Builtins.regexpmatch(product_str, expLO)
            ret = product_str
            raise Break
          end
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::GetFullProductName(%1) returned: %2",
        partial_name,
        ret
      )
      ret
    end


    # Query zypp for the list of installed products
    #
    # @return [Array<String>] installed product names
    def GetInstalledProducts
      # Query zypp for list of all known products
      product_properties = Pkg.ResolvableDependencies("", :product, "")
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(\"\", `product, \"\") returned: %1",
        product_properties
      )

      # Create the list of installed products
      installed_products = []
      Builtins.foreach(product_properties) do |product|
        if Ops.get_symbol(product, "status", :unknown) == :installed
          installed_products = Builtins.add(
            installed_products,
            Ops.get_string(product, "name", "")
          )
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::GetInstalledProducts() returned: %1",
        installed_products
      )
      deep_copy(installed_products)
    end


    # Query zypp to see if a specified product is installed.
    # For Example: IsProductInstalled("Novell-Open-Enterprise-Server") -> true if product "Novell-Open-Enterprise-Server-i386" is installed.
    #
    # @param [String] product_name name(or partial name) of product
    # @return [Boolean] true if the specified product is installed
    def IsProductInstalled(product_name)
      if product_name == nil || product_name == ""
        Builtins.y2error(
          "NovellPatterns::IsProductInstalled() returned false due to empty input parameter."
        )
        return false
      end

      # Get list of installed products
      product_properties = GetInstalledProducts()

      # Verify the specified product name is installed
      found = false
      exp = Builtins.sformat("^%1.*", product_name) # Partially match the first part of the product name. For example: "^Novell-Open-Enterprise-Server.*"
      Builtins.foreach(product_properties) do |product_str|
        if Builtins.regexpmatch(product_str, exp)
          found = true
          raise Break
        end
      end


      Builtins.y2milestone(
        "NovellPatterns::IsProductInstalled(%1) returned: %2",
        product_name,
        found
      )
      found
    end


    # Query zypp to see if Open Enterprise Server (OES) is installed
    #
    # @return [Boolean] true if OES is installed
    def IsOESInstalled
      found = IsProductInstalled(@oes_addon_name)

      Builtins.y2milestone(
        "NovellPatterns::IsOESInstalled() returned: %1",
        found
      )
      found
    end



    #FIXME: Add methods IsSLES10() and IsSLES9() here...


    # Return the detailed properties of a specified product
    #
    # @param [String] product_name name of product
    # @return [Array<Hash>] detailed properties of a given product
    def GetProductDetails(product_name)
      product_properties = []
      if product_name != nil && product_name != ""
        # Query zypp for the detailed properties of a given product
        product_properties = Pkg.ResolvableDependencies(
          product_name,
          :product,
          ""
        )
        Builtins.y2debug(
          "Pkg::ResolvableDependencies(%1, `product, \"\") returned: %2",
          product_name,
          product_properties
        )
      end

      Builtins.y2milestone(
        "NovellPatterns::GetProductDetails(\"%1\") returned: %2",
        product_name,
        product_properties
      )
      deep_copy(product_properties)
    end


    # Retrieve the name(summary) of a pattern in a
    # user readable form for display on the UI.
    #
    # @param [String] product_name name of product
    # @return [String] product name for display on the UI
    def GetProductDisplayName(product_name)
      ret = product_name
      product_details = GetProductDetails(product_name)
      Builtins.foreach(product_details) do |details|
        if Ops.get(details, "summary") != nil &&
            Ops.greater_than(
              Builtins.size(Ops.get_string(details, "summary", "")),
              0
            )
          ret = Ops.get_string(details, "summary", product_name)
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::GetProductDisplayName(%1) returned: %2",
        product_name,
        ret
      )
      ret
    end


    # Returns the source id for a given product
    #
    # @param [String] product_name name of product
    # @return [Fixnum] id of the source, -1 is returned if source is not found
    def GetProductSourceID(product_name)
      # Get the source id of this product
      product_details = GetProductDetails(product_name)
      srcid = -1
      Builtins.foreach(product_details) do |details|
        #			if (details["status"]:`unknown == `available && details["source"]:-1 != nil && details["source"]:-1 > 0) { // FIXME: sometimes details["status"] returns `selected???
        if Ops.get_integer(details, "source", -1) != nil &&
            Ops.greater_or_equal(Ops.get_integer(details, "source", -1), 0)
          repo_type = Pkg.SourceGeneralData(
            Ops.get_integer(details, "source", -1)
          )
          if Ops.get_string(repo_type, "type", "") != "YUM"
            srcid = Ops.get_integer(details, "source", -1)
            raise Break
          end
        end
      end
      Builtins.y2milestone(
        "NovellPatterns::GetProductSourceID(%1) returned: %2",
        product_name,
        srcid
      )
      srcid
    end


    # Return a list of patterns details associated with a given product.
    #
    # @param [String] product_name name of product
    # @return [Array<String>] of pattern names
    def GetProductPatterns(product_name)
      # Get the source id for this product
      srcid = GetProductSourceID(@oes_addon_name)

      # Query zypp for the detailed list of pattern dependencies
      pattern_data = Pkg.ResolvableDependencies("", :pattern, "")
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(\"\", `pattern, \"\") returned: %1",
        pattern_data
      )

      # Create a list of patterns that match the source id
      product_patterns = []
      Builtins.foreach(pattern_data) do |pattern|
        if Ops.get_integer(pattern, "source", -1) == srcid
          product_patterns = Builtins.add(
            product_patterns,
            Ops.get_string(pattern, "name", "")
          )
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::GetProductPatterns(%1) returned: %2",
        product_name,
        product_patterns
      )
      deep_copy(product_patterns)
    end


    #***********************************************************
    # Pattern query routines
    #***********************************************************


    # Query zypp for the list of installed patterns
    #
    # @return [Array<String>] installed pattern names
    def GetInstalledPatterns
      # Query zypp for a list of known patterns
      pattern_data = Pkg.ResolvableDependencies("", :pattern, "")
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(\"\", `pattern, \"\") returned: %1",
        pattern_data
      )

      # Create the list of installed patterns
      installed_patterns = []
      Builtins.foreach(pattern_data) do |pattern|
        if Ops.get_symbol(pattern, "status", :unknown) == :installed
          installed_patterns = Builtins.add(
            installed_patterns,
            Ops.get_string(pattern, "name", "")
          )
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::GetInstalledPatterns() returned: %1",
        installed_patterns
      )
      deep_copy(installed_patterns)
    end


    # Query zypp to see if a specified pattern is installed
    #
    # @param [String] pattern_name name of pattern
    # @return [Boolean] true if the specified pattern is installed
    def IsPatternInstalled(pattern_name)
      if pattern_name == nil || pattern_name == ""
        Builtins.y2error(
          "NovellPatterns::IsPatternInstalled() returned false due to empty input parameter."
        )
        return false
      end

      # Get list of pattern properties
      pattern_data = Pkg.ResolvableDependencies(pattern_name, :pattern, "")
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(%1, `pattern, \"\") returned: %2",
        pattern_name,
        pattern_data
      )
      #		y2milestone("Pkg::ResolvableDependencies(%1, `pattern, \"\") returned: %2", pattern_name, pattern_data);

      # Ensure pattern_name matches the name of installed patterns returned
      found = false
      Builtins.foreach(pattern_data) do |pattern|
        if Ops.get_string(pattern, "name", "") == pattern_name &&
            Ops.get_symbol(pattern, "status", :unknown) == :installed
          found = true
          raise Break
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::IsPatternInstalled(%1) returned: %2",
        pattern_name,
        found
      )
      found
    end


    # Generate a list of all patterns selected by zypp for installation
    #
    # @return <list>string of pattern names selected for installation
    def GetAllSelectedPatterns
      # Get list of pattern properties
      pattern_data = Pkg.ResolvableDependencies("", :pattern, "")
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(\"\", `pattern, \"\") returned: %1",
        pattern_data
      )

      # Ensure pattern_name matches the name of installed patterns returned
      selected = []
      Builtins.foreach(pattern_data) do |pattern|
        if Ops.get_symbol(pattern, "status", :unknown) == :selected
          selected = Builtins.add(selected, Ops.get_string(pattern, "name", ""))
        end
      end

      selected = Builtins.filter(selected) { |item| item != nil && item != "" } # Remove all empty list items
      selected = Builtins.toset(selected) # sort list and remove duplicates

      Builtins.y2milestone(
        "NovellPatterns::GetSelectedPatterns() returned: %1",
        selected
      )
      deep_copy(selected)
    end


    # Query zypp to see if a specified pattern is selected for installation
    #
    # @param [String] pattern_name name of pattern
    # @return [Boolean] true if the specified pattern is selected for installation
    def IsPatternSelected(pattern_name)
      if pattern_name == nil || pattern_name == ""
        Builtins.y2error(
          "NovellPatterns::IsPatternSelected() returned false due to empty input parameter."
        )
        return false
      end

      # Get list of pattern properties
      selected_patterns = GetAllSelectedPatterns()

      # Ensure pattern_name matches the name of selected patterns returned
      found = false
      Builtins.foreach(selected_patterns) do |selected|
        if selected == pattern_name
          found = true
          raise Break
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::IsPatternSelected(%1) returned: %2",
        pattern_name,
        found
      )
      found
    end


    # Return the detailed properties of a specified pattern
    #
    # @param [String] pattern_name name of pattern
    # @return [Hash] detailed properties of a given pattern
    def GetPatternDetails(pattern_name)
      if pattern_name == nil || pattern_name == ""
        Builtins.y2error(
          "NovellPatterns::GetPatternDetails() returned nil due to empty input parameter."
        )
        return nil
      end

      ret = {}
      # Query zypp for the detailed properties of a given pattern
      #		ret = Pkg::PatternData(pattern_name); // NOTE: this will not return the pattern dependencies, etc. (see comments at the top of this file)
      Builtins.y2debug("Pkg::PatternData(%1) returned: %2", pattern_name, ret)

      # Query zypp for the dependencies of a given pattern
      pattern_data_list = Pkg.ResolvableDependencies(pattern_name, :pattern, "") # NOTE: this will not return the "order" flag, etc. (see comments at the top of this file)
      Builtins.y2debug(
        "Pkg::ResolvableDependencies(%1, `pattern, \"\") returned: %2",
        pattern_name,
        pattern_data_list
      )
      Builtins.foreach(pattern_data_list) do |pattern_data|
        ret = Builtins.union(ret, pattern_data) # Merge all the pattern data together
      end

      ret = Builtins.remove(ret, "status") # NOTE: "status" flag is unreliable, use NovellPatterns::IsPatternInstalled() instead.

      ret = {} if ret == nil # don't return a nil value.
      Builtins.y2milestone(
        "NovellPatterns::GetPatternDetails(%1) returned: %2",
        pattern_name,
        ret
      )
      deep_copy(ret)
    end


    # Return the visible (user_visible) flag of a specified pattern.
    #
    # @param [String] pattern_name name of pattern
    # @return [Boolean] true if pattern is visible
    def IsPatternVisible(pattern_name)
      pattern_data = GetPatternDetails(pattern_name)
      Ops.get_boolean(pattern_data, "user_visible", false)
    end


    # Return the category flag of a specified pattern.
    #
    # @param [String] pattern_name name of pattern
    # @return [String] category
    def GetPatternCategory(pattern_name)
      pattern_data = GetPatternDetails(pattern_name)
      Ops.get_string(pattern_data, "category", "")
    end


    # Return the order flag of a specified pattern.
    #
    # @param [String] pattern_name name of pattern
    # @return interger order
    def GetPatternOrder(pattern_name)
      pattern_data = GetPatternDetails(pattern_name)
      Ops.get_string(pattern_data, "order", "")
    end


    # Retrieve the name(summary) of a pattern in a
    # user readable form for display on the UI.
    #
    # @param [String] pattern_name name of pattern
    # @return [String] pattern name for display on the UI
    def GetPatternDisplayName(pattern_name)
      pattern_data = GetPatternDetails(pattern_name)
      ret = Ops.get_string(pattern_data, "summary", "")
      if ret == nil || ret == ""
        ret = pattern_name # Avoid passing back an empty string for display on the UI.
      end
      Builtins.y2milestone(
        "NovellPatterns::GetPatternDisplayName(%1) returned: %2",
        pattern_name,
        ret
      )
      ret
    end


    # Sort a list of pattern names based on their predefined "order" (NovellPatterns::GetPatternOrder)
    #
    # @param [Array<String>] patterns_list name of patterns
    # @return [Array<String>] name of patterns sorted by order
    def SortPatternsByOrder(patterns_list)
      patterns_list = deep_copy(patterns_list)
      ret = []

      # Generate a list of maps containing the pattern name and order
      patterns_order = []
      Builtins.foreach(patterns_list) do |pattern_name|
        patterns_order = Builtins.add(
          patterns_order,
          { "name" => pattern_name, "order" => GetPatternOrder(pattern_name) }
        )
      end
      Builtins.y2milestone("patterns_order=%1", patterns_order)

      # Sort the list based on the order integer
      patterns_order = Builtins.sort(patterns_order) do |x, y|
        Ops.less_than(
          Ops.get_string(x, "order", ""),
          Ops.get_string(y, "order", "")
        )
      end # sort list in desending order (e.g. [0, 1, 2,]). NOTE: Fortunatly this builtin does not a make use of bubble sort ;-)
      Builtins.y2milestone("sorted patterns_order=%1", patterns_order)

      # Generate the sorted list of pattern names
      Builtins.foreach(patterns_order) do |pattern_data|
        ret = Builtins.add(ret, Ops.get_string(pattern_data, "name", ""))
      end

      # Remove all empty entries from the sorted patterns list
      # NOTE: we do not want to remove duplicates for various reasons, one being that the list or patterns will be resorted alphabetically. :-(
      ret = Builtins.filter(ret) { |val| val != nil && val != "" }

      ret = [] if ret == nil # avoid passing a nil value
      Builtins.y2milestone(
        "NovellPatterns::SortPatternsByOrder(%1) returned: %2",
        patterns_list,
        ret
      )
      deep_copy(ret)
    end



    #***********************************************************
    # ZYPP manipulation routines
    #***********************************************************


    # Ensure all the ZYPP database is initalized.
    # This will ensure that the pkg-builtins will work properly.
    #
    # NOTE: This method will aquire the zypp lock.
    #       This lock is a singular lock and the unlock method is not currently implemented. :-(
    #       This means that any other YaST module that requires the zypp lock will not be
    #       able to run at the same time as this module.
    #       See Bugzilla #169799
    def InitPackageDatabase
      Builtins.y2milestone(
        "NovellPatterns::Init() is initalizing the ZYPP database..."
      )
      #	Pkg::SetLocale (UI::GetLanguage (true)); // Ensure the packager knows the correct locale.
      PackageSystem.EnsureSourceInit
      PackageSystem.EnsureTargetInit # FIXME: TargetInit has recently been depricated (see pkg-builtins.changes)
      Pkg.PkgSolve(true)

      nil
    end


    # Solves all changes to the zypp database.
    # Displays pacakgemanager to solve an unresolvable(conflicting) changes
    #
    # @return [Boolean] true on success
    def SolveChanges
      success = true
      # Solve package dependencies...
      if Pkg.PkgSolve(false) == false
        Popup.ClearFeedback # clear any stray feedback messages from inst_oes_sw_select  //FIXME: should reorder when feedback messages are displayed so this doesn't happen

        Builtins.y2error(
          "Failed to solve package dependencies using Pkg::Solve()"
        )
        numErrors = Pkg.PkgSolveErrors
        Builtins.y2error("Number of pacakge conflicts: %1", numErrors)
        # FIXME: do we want to display a popup informing user of the number of package errors???
        # if ( ! Popup::YesNo(sformat(_("There are %1 package conflicts.
        #
        # Solve these conflicts now?"), numErrors))) {
        # 			success = false;
        # 		}


        # Resolve conflicts using the pacakge manager in summary mode...
        result = Convert.to_symbol(
          WFM.CallFunction("inst_packages", [:summaryMode])
        )
        Builtins.y2milestone("inst_packages returned: %1", result)
        if result != :accept && result != :next && result != :auto
          success = false
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::SolveChanges() returned: %1",
        success
      )
      success
    end


    # Commits all changes to the zypp database
    #
    # @return [Boolean] true on success
    def CommitChanges
      success = true
      Builtins.y2milestone("Commiting changes to zypp database...")
      current_control_file = ""
      if Stage.normal
        # NOTE: package manager screws with the ProductControl, as a result a ProductControl data must be saved and then restore afterwards :-@
        Yast.import "ProductControl"
        current_control_file = ProductControl.custom_control_file
        Builtins.y2milestone(
          "NovellPatterns::CommitChanges() saving control file: %1",
          current_control_file
        )
      else
        Builtins.y2milestone(
          "NovellPatterns::CommitChanges():skipping product control file manipulation since we are not Stage::normal()"
        )
      end

      # FIXME: for some reason PkgAnyToInstall and PkgAnyToDelete always return false :-(
      #	  this check is necessary since without it, the package manager slide show will always flicker before each modules configuration...
      notInstalled = GetAllSelectedPatterns()

      # novell-domain-services pattern cannot be installed / configured when eDirectory is already configured.
      Builtins.foreach(notInstalled) do |pattern|
        if pattern == "novell-domain-services"
          if FileUtils.Exists("/var/opt/novell/eDirectory/data/dib/nds.db")
            Popup.AnyMessage(
              _("Aborting DSfW Installation"),
              _(
                "eDirectory is already configured on this server. Hence, DSfW installation is aborted."
              )
            )
            success = false
            @dsfw_install_abort = true
          end
        end
      end
      # return false if novell-domain-services pattern is selected when eDirectory is already configured.
      return false if success == false

      if Pkg.IsAnyResolvable(:any, :to_install) ||
          Pkg.IsAnyResolvable(:any, :to_remove) ||
          Ops.greater_than(Builtins.size(notInstalled), 0)
        Wizard.CreateDialog # NOTE: this is necessary since the sw_single (and SuSEconfig) will mess with the current dialog causing problems later (e.g. disable the `abort button)
        Wizard.SetDesktopIcon("sw_single")

        success = success && SolveChanges()

        # Commit the package changes...
        if success
          SCR.Write(path(".target.ycp"), "/var/lib/YaST2/failed_packages", [])
          anyToDelete = Pkg.IsAnyResolvable(:any, :to_remove)
          SlideShow.SetLanguage(UI.GetLanguage(true))
          PackageSlideShow.InitPkgData(true) # force reinitialization
          SlideShow.OpenDialog

          Yast.import "PackageInstallation"
          #	Pkg::TargetLogfile (Installation::destdir + Directory::logdir + "/y2logRPM");
          oldvmlinuzsize = Convert.to_integer(
            SCR.Read(path(".target.size"), "/boot/vmlinuz")
          )
          commit_result = PackageInstallation.CommitPackages(0, 0)
          newvmlinuzsize = Convert.to_integer(
            SCR.Read(path(".target.size"), "/boot/vmlinuz")
          )

          SlideShow.CloseDialog

          if Installation.destdir == "/" &&
              (Ops.greater_than(Ops.get_integer(commit_result, 0, 0), 0) || anyToDelete)
            # prepare "you must boot" popup in inst_suseconfig
            Kernel.SetInformAboutKernelChange(oldvmlinuzsize != newvmlinuzsize)
          end
        end

        UI.CloseDialog
      else
        Builtins.y2warning(
          "WARNING: No changes to the zypp database detected. Skipping the commit."
        )
      end

      if Stage.normal
        # Restore the saved control file
        Builtins.y2milestone(
          "ProductControl::custom_control_file = %1",
          ProductControl.custom_control_file
        )
        Builtins.y2milestone(
          "NovellPatterns::CommitChanges() restoring control file: %1",
          current_control_file
        )
        ProductControl.custom_control_file = current_control_file
        if !ProductControl.Init
          # popup error message
          # %1 = absolute pathname to a product control file (i.e. '/usr/share/YaST2/control/oes_install.xml')
          Report.Error(
            Builtins.sformat(
              _("Control file %1 was not found."),
              ProductControl.custom_control_file
            )
          )
          return false
        end
      else
        Builtins.y2milestone(
          "NovellPatterns::CommitChanges():Part 2:skipping product control file manipulation since we are not Stage::normal()"
        )
      end


      Builtins.y2milestone(
        "NovellPatterns::CommitChanges() returned: %1",
        success
      )
      success
    end

    # Install a pattern and all of that pattern's dependencies (rpm packages or other patterns)
    #
    # @param [String] pattern_name name of pattern to install
    # @return [Boolean] true on success
    def InstallPattern(pattern_name)
      success = true

      # Select the pattern for installation
      success = success && Pkg.ResolvableInstall(pattern_name, :pattern)

      # Commit the changes
      success = success && CommitChanges() if success

      if !success
        Report.Error(
          Builtins.sformat(
            _("Failed to install pattern: %1"),
            GetPatternDisplayName(pattern_name)
          )
        )
      end

      Builtins.y2milestone(
        "NovellPatterns::InstallPattern() returned: %1",
        success
      )
      success
    end


    # Select a pattern and all of that pattern's dependencies for installation (rpm packages or other patterns)
    #
    # @param [String] pattern_name name of pattern to select for installation
    # @return [Boolean] true on success
    def SelectPattern(pattern_name)
      success = true

      # Select the pattern for installation
      success = success && Pkg.ResolvableInstall(pattern_name, :pattern)

      # Commit the changes
      success = success && SolveChanges() if success

      if !success
        Report.Error(
          Builtins.sformat(
            _("Failed to select pattern for installation: %1"),
            GetPatternDisplayName(pattern_name)
          )
        )
      end

      Builtins.y2milestone(
        "NovellPatterns::SelectPattern() returned: %1",
        success
      )
      success
    end


    # Set a selected patterns state to netural
    #
    # @param [String] pattern_name name of pattern to set to netural
    # @return [Boolean] true on success
    def NeturalPattern(pattern_name)
      success = true

      # De-Select the pattern for installation
      success = success && Pkg.ResolvableNeutral(pattern_name, :pattern, false)

      # Commit the changes
      success = success && SolveChanges() if success

      if !success
        Report.Error(
          Builtins.sformat(
            _("Failed to set pattern to state netural: %1"),
            GetPatternDisplayName(pattern_name)
          )
        )
      end

      Builtins.y2milestone(
        "NovellPatterns::NeturalPattern() returned: %1",
        success
      )
      success
    end


    # Using a given pattern name recursively figure out all the assosiated required
    # pattern requred dependencies
    #
    # @param [String] pattern_name name of pattern
    # @return [Array<String>] of unresolved pattern dependencies
    def GetRecursivePatternDeps(pattern_name)
      return [] if pattern_name == nil || pattern_name == ""

      pattern_deps = [pattern_name] # add the current pattern

      # recurse through the patterns....
      pattern_data = GetPatternDetails(pattern_name)
      dependencies_list = Ops.get_list(pattern_data, "dependencies", [])
      Builtins.foreach(dependencies_list) do |dep_item|
        if Ops.get_string(dep_item, "dep_kind", "") == "requires" &&
            Ops.get_string(dep_item, "res_kind", "") == "pattern"
          pattern_deps = Convert.convert(
            Builtins.union(
              pattern_deps,
              GetRecursivePatternDeps(Ops.get_string(dep_item, "name", ""))
            ),
            :from => "list",
            :to   => "list <string>"
          ) # recursively find the next set of required patterns dependencies
        end
      end

      pattern_deps = Builtins.filter(pattern_deps) do |item|
        item != nil && item != ""
      end # Remove all empty list items
      pattern_deps = Builtins.toset(pattern_deps) # sort list and remove duplicates

      deep_copy(pattern_deps)
    end


    # Using a given pattern name recursively figure out all the pattern
    # dependencies that need to be solved (missing and/or conflicting)
    #
    # @param [String] pattern_name name of pattern
    # @return [Array<String>] of unresolved pattern dependencies
    def GetUnresolvedPatternDeps(pattern_name)
      return [] if pattern_name == nil || pattern_name == ""

      Builtins.y2milestone(
        "NovellAddonInstall::GetUnresolvedPatternDeps(%1)",
        pattern_name
      )

      # Get list of pattern dependencies
      pattern_deps = GetRecursivePatternDeps(pattern_name)
      Builtins.y2debug(
        "NovellAddonInstall::GetRecursivePatternDeps returned: %1",
        pattern_deps
      )

      # Generate list of uninstalled patterns
      notInstalled = Builtins.filter(pattern_deps) do |item|
        IsPatternInstalled(item) == false
      end # Remove patterns that are already installed

      notInstalled = Builtins.filter(notInstalled) do |item|
        item != nil && item != ""
      end # Remove all empty list items
      notInstalled = Builtins.toset(notInstalled) # sort list and remove duplicates

      Builtins.y2milestone(
        "NovellPatterns::GetUnresolvedPatternDeps returned: %1",
        notInstalled
      )
      deep_copy(notInstalled)
    end


    # If no patterns are selected for installation, returns true w/o any UI
    # Otherwise, if any pattern is selected for installation then
    # display a popup confirming these changes before preforming the installation of the patterns.
    #
    # @return [Boolean] true if the pattern changes are valid
    def ConfirmPatternChanges
      success = true

      # Get list of patterns selected for installation
      notInstalled = GetAllSelectedPatterns()

      # Display pop-up for any required patterns that are not installed
      if Ops.greater_than(Builtins.size(notInstalled), 0)
        # Generate a list of user readable pattern names to display in the Popup.
        display_names_list = []
        Builtins.foreach(notInstalled) do |item|
          if IsPatternVisible(item)
            display_names_list = Builtins.add(
              display_names_list,
              GetPatternDisplayName(item)
            )
          end
        end
        display_names_list = Builtins.filter(display_names_list) do |item|
          item != nil && item != ""
        end # Remove all empty list items

        # Display a Yes/No popup
        if Ops.greater_than(Builtins.size(display_names_list), 0)
          solve_dep = Popup.AnyQuestionRichText(
            _("Unresolved Product Dependencies"),
            # rich text message
            # %1 product neme with the unresolved dependencies
            # %2 list of product names
            Builtins.sformat(
              _(
                "<p>The selected items can not be configured until the following patterns are installed:</p><p><b>%1</b></p><p>Do you want to install them now?</p>"
              ),
              Builtins.mergestring(display_names_list, "<br>")
            ),
            50,
            15,
            # popup message button
            _("&Install Now..."),
            Label.CancelButton,
            :yes
          )

          if !solve_dep
            success = false # `no button was selected on the Popup
          end
        end

        # Install any required patterns that an not installed
        success = success && CommitChanges() if success
      end

      Builtins.y2milestone(
        "NovellPatterns::ConfirmPatternChanges returned: %1",
        success
      )
      success
    end


    # Ensure that a specified list of patterns are installed or selected for
    # If the pattern is installed returns true w/o any UI
    # Otherwise, if the pattern is not installed then
    # display a popup asking to install the pattern.
    #
    # @param [Array<String>] patterns_list list<string> of pattern names
    # @return [Boolean] true if the pattern is installed
    def EnsurePatternsInstalled(patterns_list)
      patterns_list = deep_copy(patterns_list)
      Builtins.y2milestone(
        "NovellPatterns::EnsurePatternInstalled(%1)",
        patterns_list
      )
      success = true

      # select each pattern for installation if it is not already installed
      Builtins.foreach(patterns_list) do |pattern_name|
        if !IsPatternInstalled(pattern_name) && !IsPatternSelected(pattern_name)
          # Select the pattern for installation
          success = success && Pkg.ResolvableInstall(pattern_name, :pattern)
        end
      end

      success = success && ConfirmPatternChanges()
      if success == false
        # undo the pattern selections
        Builtins.foreach(patterns_list) do |pattern_name|
          NeturalPattern(pattern_name) if IsPatternSelected(pattern_name)
        end
      end

      Builtins.y2milestone(
        "NovellPatterns::EnsurePatternInstalled returned: %1",
        success
      )
      success
    end



    # WARNING: this method is deprecated
    #
    # FIXME: must remove this method when all dependent modules have been updated to use EnsurePatternsInstalled(list<string>)
    def EnsurePatternInstalled(pattern_name)
      Builtins.y2error(
        "WARNING: NovellPatterns::EnsurePatternInstalled(string).  Please use EnsurePatternsInstalled(list<string>) instead."
      )
      EnsurePatternsInstalled([pattern_name])
    end



    # Remove a pattern and all of that pattern's dependencies (rpm packages or other patterns)
    #
    # @param [String] pattern_name name of pattern to remove
    # @return [Boolean] true on success
    def RemovePattern(pattern_name)
      success = true

      # NOTE: removing patterns is different...successful removal will require a recursive algorithm....// FIXME: how do we ensure we're not recursivly removing things we shouldn't remove???

      # Remove the rpm packages assosiated with this pattern
      # FIXME:  must add list of rpm packages here!
      #success = success && PackageSystem::DoRemove(FIXME:???);


      # FIXME: must recursivly uninstall the associated patterns here!
      # FIXME: what is the desired behavior...do we really want to automatically uninstall every pattern?


      # Select the pattern for installation
      success = success && Pkg.ResolvableRemove(pattern_name, :pattern)

      success = success && CommitChanges() if success

      if !success
        Report.Error(
          Builtins.sformat(_("Failed to remove pattern: %1"), pattern_name)
        )
      end

      Builtins.y2milestone(
        "NovellPatterns::RemovePattern returned: %1",
        success
      )
      success
    end


    # Remove (un-install) an rpm package
    #
    # @param [String] package_name name of package to remove
    # @return [Boolean] true on success
    def RemovePackage(package_name)
      success = true

      # Select the package for removal
      success = success && Pkg.ResolvableRemove(package_name, :package)

      # Commit the changes
      success = success && CommitChanges() if success

      if !success
        Report.Error(
          Builtins.sformat(_("Failed to remove rpm package: %1"), package_name)
        )
      end

      Builtins.y2milestone(
        "NovellPatterns::RemovePackage() returned: %1",
        success
      )
      success
    end


    #Function to Disabled enbale multiverionkernel in /mnt/etc/zypp/zypp.conf or /etc/zypp/zypp.conf

    def DisableEnableMultiVersionKernelZyppConf(path)

      zypp_conf_location=path

      if zypp_conf_location == "/mnt/etc/zypp/zypp.conf"
        cmd = "grep '^multiversion = provides:multiversion(kernel)' #{zypp_conf_location}"
        rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
        Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        if Ops.get_integer(rval, "exit", -1) == 0
          cmd = "sed -i 's/multiversion = provides:multiversion(kernel)/#multiversion = provides:multiversion(kernel)/g' #{zypp_conf_location}"
          rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
          Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        end

        cmd = "grep '^multiversion.kernels = latest,latest-1,running' #{zypp_conf_location}"
        rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
        Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        if Ops.get_integer(rval, "exit", -1) == 0
          cmd = "sed -i 's/multiversion.kernels = latest,latest-1,running/#multiversion.kernels = latest,latest-1,running/g' #{zypp_conf_location}"
          rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
          Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        end
      elsif zypp_conf_location == "/etc/zypp/zypp.conf"
        cmd = "grep '#multiversion = provides:multiversion(kernel)' #{zypp_conf_location} |grep \"kernel)$\""
        rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
        Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        if Ops.get_integer(rval, "exit", -1) == 0
          cmd="sed -i  's/#multiversion = provides:multiversion(kernel)/multiversion = provides:multiversion(kernel)/g' #{zypp_conf_location}"
          rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
          Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        end
        cmd = "grep '#multiversion.kernels = latest,latest-1,running' #{zypp_conf_location}"
        rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
        Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        if Ops.get_integer(rval, "exit", -1) == 0
          cmd = "sed -i 's/#multiversion.kernels = latest,latest-1,running/multiversion.kernels = latest,latest-1,running/g' #{zypp_conf_location}"
          rval = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
          Builtins.y2milestone("cmd is %1, rval is %2",cmd,rval)
        end
      end
      nil
    end


    # Constructor - Initializes zypp database and defaults the full OES product name as it is know by zypp.
    # Ensures that Pkg:: calls will work with the installation sources.
    #
    # @return [void]
    def NovellPatterns

      # initialize the zypp database.
      InitPackageDatabase()

      # Replace the partial name of the OES product with it's full name  For example: "Open_Enterprise_Server" or "Novell-Open-Enterprise-Server-i386"
      @oes_addon_name = GetFullProductName(".*Open.*Enterprise.*Server.*") # regexp for the product name
      Builtins.y2milestone(
        "NovellPatterns::oes_addon_name = %1",
        @oes_addon_name
      )

      nil
    end

    publish :variable => :oes_addon_name, :type => "string"
    publish :variable => :dsfw_install_abort, :type => "boolean"
    publish :function => :GetAllKnownProducts, :type => "list <string> ()"
    publish :function => :GetFullProductName, :type => "string (string)"
    publish :function => :GetInstalledProducts, :type => "list <string> ()"
    publish :function => :IsProductInstalled, :type => "boolean (string)"
    publish :function => :IsOESInstalled, :type => "boolean ()"
    publish :function => :GetProductDetails, :type => "list <map> (string)"
    publish :function => :GetProductDisplayName, :type => "string (string)"
    publish :function => :GetProductSourceID, :type => "integer (string)"
    publish :function => :GetProductPatterns, :type => "list <string> (string)"
    publish :function => :GetInstalledPatterns, :type => "list <string> ()"
    publish :function => :IsPatternInstalled, :type => "boolean (string)"
    publish :function => :GetAllSelectedPatterns, :type => "list <string> ()"
    publish :function => :IsPatternSelected, :type => "boolean (string)"
    publish :function => :GetPatternDetails, :type => "map (string)"
    publish :function => :IsPatternVisible, :type => "boolean (string)"
    publish :function => :GetPatternCategory, :type => "string (string)"
    publish :function => :GetPatternOrder, :type => "string (string)"
    publish :function => :GetPatternDisplayName, :type => "string (string)"
    publish :function => :SortPatternsByOrder, :type => "list <string> (list <string>)"
    publish :function => :InitPackageDatabase, :type => "void ()"
    publish :function => :SolveChanges, :type => "boolean ()"
    publish :function => :CommitChanges, :type => "boolean ()"
    publish :function => :InstallPattern, :type => "boolean (string)"
    publish :function => :SelectPattern, :type => "boolean (string)"
    publish :function => :NeturalPattern, :type => "boolean (string)"
    publish :function => :GetRecursivePatternDeps, :type => "list <string> (string)"
    publish :function => :GetUnresolvedPatternDeps, :type => "list <string> (string)"
    publish :function => :ConfirmPatternChanges, :type => "boolean ()"
    publish :function => :EnsurePatternsInstalled, :type => "boolean (list <string>)"
    publish :function => :EnsurePatternInstalled, :type => "boolean (string)"
    publish :function => :RemovePattern, :type => "boolean (string)"
    publish :function => :RemovePackage, :type => "boolean (string)"
    publish :function => :NovellPatterns, :type => "void ()"
    publish :function => :DisableEnableMultiVersionKernelZyppConf, :type => "void ()"
  end

  NovellPatterns = NovellPatternsClass.new
  NovellPatterns.main
end
