# encoding: utf-8

# Copyright (c) 2020 Micro Focus or one of its affiliates.
# File:	modules/NovellValidators.ycp
# Package:	novell-common
# Summary:	Novell validation routines
# Authors:	Stephen T. Bunch <sbunch@novell.com>
#
# $Id: NovellValidators.ycp,v 1.1.2.19 2005/09/06 19:04:00 rhaslam Exp $
#
require "yast"

module Yast
  class NovellValidatorsClass < Module
    def main

      textdomain "novell-base"

      Yast.import "NovellUtils"
      Yast.import "Report"
      Yast.import "Popup"
      Yast.import "IP"
      Yast.import "Address"
      Yast.import "String"
      Yast.import "ValidationHelpers"

      @invalid_context = _("%1 is an invalid context")
      @invalid_state_context = _(
        "Invalid 'State' context type. You have entered a context that specifies a 'state' attribute (e.g. s=alaska or st=alaska).\n" +
          "LDAP (comma separated) format requires that the state attribute be specified as 'st' (e.g. st=alaska).\n" +
          "NLDAP (period separated) format requires that the state attribute be specified as 's' (e.g. s=alaska).\n" +
          "For example, 'State' context in ldap format is 'o=novell,l=milan,st=berlin,c=gr'.\n" +
          "'State' context in NDS format is 'o=novell.l=milan.s=berlin.c=gr'."
      )
    end

    # Check length of context for 256 maximum and warn if too close
    def CheckContextLength(context)
      ok = true

      sizeOfContext = Builtins.size(context)
      if Ops.greater_or_equal(sizeOfContext, 226) # warn the user about approaching the NDS context size limit of 256 characters.
        contextSizeWarningUnder = _(
          "The size of the context entered is approaching the NDS limit of 256 characters.\n" +
            "Internal system additions to this context (if any) may cause the size\n" +
            "of the context to exceed the 256 limit.\n" +
            "Do you want to continue anyway?\n" +
            "\n" +
            "Context size: %1\n" +
            "Context: '%2'"
        )
        contextSizeWarningOver = _(
          "The size of the context entered has exceeded the NDS limit of 256 characters.\n" +
            "\n" +
            "Context size: %1\n" +
            "Context: '%2'"
        )

        allowContinue = true
        allowContinue = false if Ops.greater_than(sizeOfContext, 256)

        if allowContinue
          goOn = Popup.ContinueCancelHeadline(
            _("NDS Context Size Warning"),
            Builtins.sformat(contextSizeWarningUnder, sizeOfContext, context)
          )
          ok = false if !goOn
        else
          Popup.Error(
            Builtins.sformat(contextSizeWarningOver, sizeOfContext, context)
          )
          ok = false
        end
      end
      ok
    end

    # See if the state context type is incorrect for ldap/nds format
    def WrongState(name)
      retVal = false

      if NovellUtils.isLDAP(name) # ldap format
        # Check for invalid state attribute
        retVal = true if Builtins.issubstring(Builtins.tolower(name), "s=") # NDS format
      else
        # Check for invalid state attribute
        retVal = true if Builtins.issubstring(Builtins.tolower(name), "st=")
      end
      retVal
    end

    # Check tree name
    #
    # NOTE: This code assumes that VALID TREE Names contain only [A-Z],[a-z] and "-" or "_"
    #
    # @param string tree name
    # @return [Boolean] true if the tree nam eis valid, else false
    def ValidateTreeName(treeName)
      treeNameUpper = Builtins.toupper(treeName)
      if Builtins.size(treeName) == 0
        Report.Error(_("Cannot have an empty tree name"))
        return false
      elsif Ops.greater_than(Builtins.size(treeName), 32)
        Report.Error(_("Tree name cannot be longer than 32 characters"))
        return false
      elsif Builtins.findlastof(treeName, "\\") ==
          Ops.subtract(Builtins.size(treeName), 1)
        # FIXEME
        Report.Error(_("Tree name cannot end in a backslash (\"\\\")"))
        return false
      elsif Builtins.findfirstnotof(
          treeNameUpper,
          "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-_"
        ) != nil
        Report.Error(
          _(
            "Tree names may only contain the following characters [A-Z], [a-z], [0-9] and \"-\" or \"_\""
          )
        )
        return false
      else
        return true
      end
    end


    # Validates two passwords to ensure they are the same and do not contain any invalid chars
    # @param string first password
    # @param string second password
    # @return [Boolean] true if the password is valid
    def ValidatePasswords(pw1, pw2)
      valid_password_chars = "-0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!@\#$%^&*~(),;:._+/|? {}=[]"
      if pw1 == nil || pw1 == "" || pw2 == nil || pw2 == ""
        Report.Error(_("The password is missing.\nPlease try again."))
        return false
      elsif pw1 != pw2
        Report.Error(
          _(
            "The first and the second version\n" +
              "of the password don't match!\n" +
              "Please try again."
          )
        )
        return false
      elsif Builtins.findfirstnotof(pw1, valid_password_chars) != nil
        pos = Builtins.findfirstnotof(pw1, valid_password_chars)


        Report.Error(
          _(
            "The password may only contain the following characters:\n" +
              " 0..9, a..z, A..Z, and any of \"#*,.;:._-+~!$%&/|?{[()]}\".\n" +
              "Please try again."
          )
        )
        return false
      end

      true
    end


    # Validate a password
    # @param string password
    # @return [Boolean] true if the password is valid
    def ValidatePassword(pw1)
      ValidatePasswords(pw1, pw1)
    end


    # Validate port, is it a valid number and is it already in use.
    # Displays an warning to the user if the port is in use.
    # @param [Fixnum] port number
    # @return [Boolean] true if the port is valid, else false
    def ValidatePort(port)
      if Ops.less_than(port, 0) || Ops.greater_than(port, 65536)
        Report.Error(
          _("Valid port numbers are from 0 to 65536.\nPlease try again.")
        )
        return false
      end
      ret = Convert.to_integer(
        SCR.Execute(
          path(".target.bash"),
          Builtins.sformat("ss -tulpn | grep \":%1 \"", port)
        )
      )
      if ret == 0
        if !Popup.YesNo(
            Builtins.sformat(_("Port %1 is in use, use anyway?"), port)
          )
          return false
        end
      end
      true
    end


    # Validate FDN. If the name is optional pass in 'true' as the second
    # param, otherwise pass in 'false'.  In some cases we allow an optional
    # name, in others the name is manditory.
    #
    # @param string FDN name to validate
    # @param [Boolean] optional?
    # @return true if the fully distiguished name is valid, else false
    def ValidateFDN(name, optional)
      orgContext = name
      Builtins.y2debug("Input to ValidateFDN: %1", name)
      retVal = true
      name = NovellUtils.Ldap2Nds(name)
      name = NovellUtils.Strip(name, " ")
      if name == _("none") || name == "" || name == " "
        retVal = false if !optional
      else
	name = Builtins.tolower(name)
	split = Builtins.splitstring(name, ".")

        Builtins.foreach(split) do |testLine|
          if Builtins.issubstring(testLine, "o=")
           if(testLine == "o=security")
             Report.Error(
               _(
                 "The container 'o=security' is reserved and cannot be used in a fully distinguished name"
                )
              )
              retVal = false
           elsif !ValidationHelpers.ValidateFDN(name)
              Report.Error(
              Builtins.sformat(
                _("%1 is an invalid fully distinguished name"),
                name
                )
              )
              retVal = false
           end
          end
        end 
        # if there is a country element (e.g. c=us), make sure it only has two characters.
        if retVal == true
          name = NovellUtils.Ldap2Nds(name)
          name = Builtins.tolower(name)
          split = Builtins.splitstring(name, ".")

          Builtins.foreach(split) do |testLine|
            if Builtins.issubstring(testLine, "c=")
              split2 = Builtins.splitstring(testLine, "=")
              if Ops.get(split2, 0, "") == "c"
                if Ops.greater_than(Builtins.size(Ops.get(split2, 1, "")), 2)
                  Report.Error(
                    _(
                      "The country code may only contain two characters (e.g. c=us)"
                    )
                  )
                  retVal = false
                  raise Break
                end
              end
            end
          end

          # Make sure we are using the correct State attribute ("st" or "s") depending on ldap or nldap
          if WrongState(orgContext)
            Report.Error(@invalid_state_context)
            retVal = false
          end
          retVal = false if CheckContextLength(orgContext) == false
        end
      end
      retVal
    end

    # Validate Context Exists in the tree
    # @param string context to validate
    # @return [Boolean] true on success
    def ValidateContextExists(contextName, _IP, ldapPort, admin, password)
      retCode = true
      Builtins.y2milestone(
        "NovellValidators.ValidateContextExists:Context to search for = %1",
        contextName
      )

      # ldapsearch -H ldaps://137.65.59.48:636 -D cn=Administrator,cn=users,dc=brett,dc=com -w novell -x -b ou=novell,dc=brett,dc=com -s base
      cmd = Builtins.sformat(
        "ldapsearch -H ldaps://%1:%2 -D %3 -w '%4' -x -b %5 -s base",
        _IP,
        ldapPort,
        NovellUtils.Nds2Ldap(admin),
        password,
        NovellUtils.Nds2Ldap(contextName)
      )
      Builtins.y2milestone(
        "NovellValidators.ValidateContextExists:executing %1",
        Builtins.sformat(
          "ldapsearch -H ldaps://%1:%2 -D %3 -w %4 -x -b %5 -s base",
          _IP,
          ldapPort,
          NovellUtils.Nds2Ldap(admin),
          "****",
          NovellUtils.Nds2Ldap(contextName)
        )
      )
      ccode = Convert.to_integer(SCR.Execute(path(".target.bash"), cmd))
      retCode = false if ccode != 0 # the context exists

      Builtins.y2milestone(
        "NovellValidators.ValidateContextExists:retCode %1",
        retCode
      )
      retCode
    end

    # Validate CN Exists in the tree
    # @param string context to validate
    # @return [Boolean] true on success
    def ValidateCNExists(cnName, _IP, ldapPort, admin, password)
      retCode = true
      Builtins.y2milestone(
        "NovellValidators.ValidateCNExists:CN to search for = %1",
        cnName
      )

      # ldapsearch -H ldaps://137.65.59.48:636 -D cn=Administrator,cn=users,dc=brett,dc=com -w novell -x -b ou=novell,dc=brett,dc=com -s base
      cmd = Builtins.sformat(
        "ldapsearch -H ldaps://%1:%2 -D %3 -w '%4' -x -b \"%5\" -s base 1.1",
        _IP,
        ldapPort,
        NovellUtils.Nds2Ldap(admin),
        password,
        NovellUtils.Nds2Ldap(cnName)
      )
      Builtins.y2milestone(
        "NovellValidators.ValidateCNExists:executing %1",
        Builtins.sformat(
          "ldapsearch -H ldaps://%1:%2 -D %3 -w %4 -x -b %5 -s base",
          _IP,
          ldapPort,
          NovellUtils.Nds2Ldap(admin),
          "****",
          NovellUtils.Nds2Ldap(cnName)
        )
      )
      ccode = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
      exit = Ops.get_integer(ccode, "exit", -1)
      retCode = false if exit != 0 # the context exists

      Builtins.y2milestone(
        "NovellValidators.ValidateCNExists:retCode %1",
        retCode
      )
      retCode
    end

    # Validate Context
    # @param string context to validate
    # @return [Boolean] true on success
    def ValidateContext(name)
      Builtins.y2milestone("Source context to evaluate = %1", name)
      orgContext = name
      retVal = true
      name = NovellUtils.Ldap2Nds(name)
      name = NovellUtils.Strip(name, " ")
      name = Builtins.tolower(name)
      split = Builtins.splitstring(name, ".")
      
      Builtins.foreach(split) do |testLine|
        if Builtins.issubstring(testLine, "o=")
         if(testLine == "o=security")
           Report.Error(
             _(
               "The container 'o=security' is reserved and cannot be used in a context"
              )
           )
           retVal = false
         elsif !ValidationHelpers.ValidateContext(name)
            Report.Error(Builtins.sformat(@invalid_context, name))
            retVal = false
         end
        end
      end
      # if there is a country element (e.g. c=us), make sure it only has two characters.
      if retVal == true
        #name = NovellUtils.Ldap2Nds(name)
        #name = Builtins.tolower(name)
        #split = Builtins.splitstring(name, ".")

        Builtins.foreach(split) do |testLine|
          if Builtins.issubstring(testLine, "c=")
            split2 = Builtins.splitstring(testLine, "=")
            if Ops.get(split2, 0, "") == "c"
              if Ops.greater_than(Builtins.size(Ops.get(split2, 1, "")), 2)
                Report.Error(
                  _(
                    "The country code may only contain two characters (e.g. c=us)"
                  )
                )
                retVal = false
                raise Break
              end
            end
          end
        end
        # Make sure we are using the correct State attribute ("st" or "s") depending on ldap or nldap
        if WrongState(orgContext)
          Report.Error(@invalid_state_context)
          retVal = false
        end
        retVal = false if CheckContextLength(orgContext) == false
      end
      retVal
    end






    # Validate credentials
    # @param [String] userName   - valid user name
    # @param [String] password   - password for user
    # @param stirng IP address - ip address to direct LDAP bind to
    # @param [Fixnum] port      - LDAP port to direct requrest to
    # @param [Boolean] doFeedback - does the caller provide feedback (false) or should this function provide feedback (true)
    # @return [Fixnum] 0 on success, otherwise a non-zero error code
    def ValidateCredentials(userName, password, ip, port, doFeedback)
      Builtins.y2milestone(
        "Credential info to validate: '%1', '%2', %3, %4, %5",
        userName,
        "*****",
        ip,
        port,
        doFeedback
      )
      exit = 0
      if doFeedback
        Popup.ShowFeedback(_("Verifying user credentials"), _("Please wait..."))
      end
      valid = ValidationHelpers.ValidateCredentials(
        userName,
        password,
        ip,
        port
      )
      Popup.ClearFeedback if doFeedback
      if valid != "1"
        Report.Error(
          Builtins.sformat(
            _(
              "Authentication failed because of one of the following:\n"+
              "- Server is not reachable\n"+
              "- eDirectory is not up\n"+
              "- Incorrect credentials\n"+
              "- Certificate is expired"
            )
          )
        )
        error_code = Builtins.splitstring(valid, "()")
        return Builtins.tointeger(Ops.get(error_code, 1, ""))
      else
        return 0
      end
    end





    # Validate credentials
    # @param [String] userName   - valid user name
    # @param [String] password   - password for user
    # @param stirng IP address - ip address to direct LDAP bind to
    # @param [Fixnum] port      - LDAP port to direct requrest to
    # @param [Boolean] doFeedback - does the caller provide feedback (false) or should this function provide feedback (true)
    # @return [Fixnum] 0 on success, otherwise a non-zero error code
    def ValidateCredentialsSamba(userName, password, ip, port, doFeedback)
      Builtins.y2milestone(
        "Credential info to validate: '%1', '%2', %3, %4, %5",
        userName,
        "*****",
        ip,
        port,
        doFeedback
      )
      exit = 0
      if doFeedback
        Popup.ShowFeedback(_("Verifying user credentials"), _("Please wait..."))
      end
      valid = ValidationHelpers.ValidateCredentialsSamba(
        userName,
        password,
        ip,
        port
      )
      Popup.ClearFeedback if doFeedback

      exit = 1 if !valid
      exit
    end


    # Validate IP address
    # @param IP address to check
    # @param string message to dispay if IP address is not valid
    # @return [Boolean] true if valid IP address
    def ValidateIPMsg(ip, msg)
      if !IP.Check4(ip)
        if ip != ""
          Popup.Error(IP.Valid4)
        else
          Popup.Error(msg)
        end
        return false
      end
      true
    end


    # Validate an IP address
    # @param string IP address
    # @return [Boolean] true if valid IP address
    def ValidateIP(ip)
      ValidateIPMsg(ip, _("Enter the existing server IP address."))
    end


    # Validate DNS or IP address
    # @param string IP or dns address
    # @return [Boolean] true if valid DNS or IP address
    def ValidateDNSorIP(ip)
      if !Address.Check4(ip)
        Popup.Error(Address.Valid4)
        return false
      end
      true
    end


    # Returns true is the string is a valid device
    # @param [String] device
    # @return booelan
    def ValidateDevice(device)
      device = NovellUtils.Strip(device, " \t")
      return true if device == ""

      #check to see if the path exists
      rmap = Convert.to_map(SCR.Read(path(".target.lstat"), device))
      Builtins.y2debug("stat '%1' = %2", device, rmap)
      if rmap == {}
        Builtins.y2error("Error: The path '%1' does not exist", device)
        return false
      end
      #check to see if the device is valid(S_ISBLK) and return the result
      Ops.get_boolean(rmap, "isblock", false)
    end


    # Execute the ntpdate command to see if a give server responds to NTP requests
    # @param string IP address or hostname of NTP server
    # @return [Boolean] true on success
    def ValidateNTPServer(server)
      cmd = Builtins.sformat("/usr/sbin/ntpdate -u %1", server)
      #Execute the NTP date command and look for a valid response
      rmap = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
      if Ops.get_integer(rmap, "exit", -1) != 0
        Builtins.y2milestone(
          "The time server %1 did not respond to ntpdate",
          server
        )
        return false
      else
        return true
      end
    end


    # ValidatePathname - Validate the characters in a proposed pathname
    # NOTE: May not work for double-byte languages???
    #
    # @param [String] pathname to validate
    # @param [Boolean] absolute - if true validate that pathname is absolute pathname
    # @return [Boolean] true on success
    def ValidatePathname(pathname, absolute)
      pathnameValidChars = Ops.add(String.ValidCharsFilename, "/")

      invalidChars = Builtins.deletechars(pathname, pathnameValidChars)
      return false if Builtins.size(invalidChars) != 0

      # disallow //
      if Builtins.search(pathname, "//") != nil
        Report.Error(
          Builtins.sformat(_("%1 is not a valid pathname"), pathname)
        )
        return false
      end
      if !absolute
        return true
      else
        slashloc = Builtins.findfirstof(pathname, "/")
        if slashloc != 0
          Report.Error(
            Builtins.sformat(_("%1 is not a valid absolute pathname"), pathname)
          )
          return false
        end
      end
      true
    end

    publish :variable => :invalid_context, :type => "string"
    publish :variable => :invalid_state_context, :type => "string"
    publish :function => :CheckContextLength, :type => "boolean (string)"
    publish :function => :ValidateTreeName, :type => "boolean (string)"
    publish :function => :ValidatePasswords, :type => "boolean (string, string)"
    publish :function => :ValidatePassword, :type => "boolean (string)"
    publish :function => :ValidatePort, :type => "boolean (integer)"
    publish :function => :ValidateFDN, :type => "boolean (string, boolean)"
    publish :function => :ValidateContextExists, :type => "boolean (string, string, string, string, string)"
    publish :function => :ValidateCNExists, :type => "boolean (string, string, string, string, string)"
    publish :function => :ValidateContext, :type => "boolean (string)"
    publish :function => :ValidateCredentials, :type => "integer (string, string, string, integer, boolean)"
    publish :function => :ValidateCredentialsSamba, :type => "integer (string, string, string, integer, boolean)"
    publish :function => :ValidateIPMsg, :type => "boolean (string, string)"
    publish :function => :ValidateIP, :type => "boolean (string)"
    publish :function => :ValidateDNSorIP, :type => "boolean (string)"
    publish :function => :ValidateDevice, :type => "boolean (string)"
    publish :function => :ValidateNTPServer, :type => "boolean (string)"
    publish :function => :ValidatePathname, :type => "boolean (string, boolean)"
  end

  NovellValidators = NovellValidatorsClass.new
  NovellValidators.main
end
