# encoding: utf-8

# File:
#	OesAddOnProduct.ycp
#
# Module:
#	OesAddOnProduct
#
# Summary:
#	This module provides integration of the add-on products
#
# Authors:
#	Jiri Srain<jsrain@suse.de>
#      Mark Clark - modified version for OES 2 SP2 functionality
require "yast"

module Yast
  class OesAddOnProductClass < Module
    def main
      Yast.import "Pkg"
      Yast.import "UI"

      textdomain "novell-base"

      Yast.import "Label"
      Yast.import "Mode"
      Yast.import "ProductControl"
      Yast.import "ProductFeatures"
      Yast.import "Report"
      Yast.import "XML"
      Yast.import "Wizard"
      Yast.import "FileUtils"
      Yast.import "Language"
      Yast.import "Popup"
      Yast.import "InstShowInfo"
      Yast.import "ProductLicense"
      Yast.import "Directory"
      Yast.import "Installation"
      Yast.import "String"
      Yast.import "NovellUtils"
      Yast.import "Stage"

      # variables for installation with product
      # ID for cache in the inst-sys
      @src_cache_id = -1

      # System proposals have already been prepared for merging?
      @system_proposals_prepared = false

      # System workflows have already been prepared for merging?
      @system_workflows_prepared = false

      # List of all selected sources
      #
      #
      # **Structure:**
      #
      #     add_on_products = [
      #        $[
      #          "media" : 4, // ID of the source
      #          "product_dir" : "/",
      #          "product" : "openSUSE version XX.Y",
      #        ],
      #        ...
      #      ]
      @add_on_products = []

      # ID of currently added source for the add-on product
      @src_id = nil

      # for the add-on product workflow - needed for dialog skipping
      # return value of last step in the product adding workflow
      @last_ret = nil

      # List of used control files, to be copied to target system
      @control_files_to_add = []

      # Items if add-on product to be performed at the start of inst_finish
      @finish_steps_before_chroot = []

      # Items of add-on product to be performed after swictching to chroot
      @finish_steps_after_chroot = []

      # Items of add-on product to be performed before unmounting disks
      @finish_steps_before_umount = []

      @modified = false

      @mode_config_sources = []

      # --> FATE #302123: Allow relative paths in "add_on_products" file
      @base_product_url = nil

      @initial_workflows = []
      @initial_proposals = []
      @initial_inst_finish = []
      @initial_clone_modules = []

      @initial_settings_stored = false

      # Contains list of source IDs that request registration
      @addons_requesting_registration = []

      # Every Add-On can preselect some patterns.
      # Only patterns that are not selected/installed yet will be used.
      #
      #
      # **Structure:**
      #
      #     $[
      #        src_id : [
      #          "pattern_1", "pattern_2", "pattern_6"
      #        ]
      #      ]
      @patterns_preselected_by_addon = {}
    end

    def SetBaseProductURL(url)
      Builtins.y2warning("Empty base url") if url == "" || url == nil

      @base_product_url = url
      Builtins.y2milestone("New base URL: %1", @base_product_url)

      nil
    end

    def GetBaseProductURL
      @base_product_url
    end

    # Reads list of HTTP params and returns them as map
    #
    # @param [String] params
    # @return [Hash{String => String}] params
    #
    # @example
    #      MakeMapFromParams ("device=sda3&login=aaa&password=bbb") -> $[
    #              "device"   : "sda3",
    #              "login"    : "aaa",
    #              "password" : "bbb"
    #      ]
    def MakeMapFromParams(params)
      params_list = Builtins.splitstring(params, "&")

      params_list = Builtins.filter(params_list) do |one_param|
        one_param != "" && one_param != nil
      end

      ret = {}
      eq_pos = nil
      opt = ""
      val = ""

      Builtins.foreach(params_list) do |one_param|
        eq_pos = Builtins.search(one_param, "=")
        if eq_pos == nil
          Ops.set(ret, one_param, "")
        else
          opt = Builtins.substring(one_param, 0, eq_pos)
          val = Builtins.substring(one_param, Ops.add(eq_pos, 1))

          Ops.set(ret, opt, val)
        end
      end

      deep_copy(ret)
    end

    # Returns an absolute URL from base + relative url.
    # Relative URL needs to start with 'reulrl://' othewise
    # it is not considered being relative and it's returned
    # as it is (just the relative_url parameter).
    #
    # @param [String] base_url
    # @param string relative_url
    # @return [String] absolute_url
    #
    # @example
    #   OesAddOnProduct::GetAbsoluteURL (
    #     "http://www.example.org/some%20dir/another%20dir",
    #     "relurl://../AnotherProduct/"
    #   ) -> "http://www.example.org/some%20dir/AnotherProduct/"
    #   OesAddOnProduct::GetAbsoluteURL (
    #     "username:password@ftp://www.example.org/dir/",
    #     "relurl://./Product_CD1/"
    #   ) -> "username:password@ftp://www.example.org/dir/Product_CD1/"
    def GetAbsoluteURL(base_url, url)
      if !Builtins.regexpmatch(url, "^relurl://")
        Builtins.y2debug("Not a relative URL: %1", url)
        return url
      end

      if base_url == nil || base_url == ""
        Builtins.y2error("No base_url defined")
        return url
      end

      # bugzilla #306670
      base_params_pos = Builtins.search(base_url, "?")
      base_params = ""

      if base_params_pos != nil && Ops.greater_or_equal(base_params_pos, 0)
        base_params = Builtins.substring(base_url, Ops.add(base_params_pos, 1))
        base_url = Builtins.substring(base_url, 0, base_params_pos)
      end

      added_params_pos = Builtins.search(url, "?")
      added_params = ""

      if added_params_pos != nil && Ops.greater_or_equal(added_params_pos, 0)
        added_params = Builtins.substring(url, Ops.add(added_params_pos, 1))
        url = Builtins.substring(url, 0, added_params_pos)
      end

      base_url = Ops.add(base_url, "/") if !Builtins.regexpmatch(base_url, "/$")

      Builtins.y2milestone(
        "Merging '%1' (params '%2') to '%3' (params '%4')",
        url,
        added_params,
        base_url,
        base_params
      )
      url = Builtins.regexpsub(url, "^relurl://(.*)$", "\\1")

      url = Builtins.sformat("%1%2", base_url, url)

      # merge /something/../
      max_count = 100

      while Ops.greater_than(max_count, 0) &&
          Builtins.regexpmatch(url, "(.*/)[^/]+/+\\.\\./")
        max_count = Ops.subtract(max_count, 1)
        str_offset_l = Builtins.regexppos(url, "/\\.\\./")
        str_offset = Ops.get(str_offset_l, 0)

        if str_offset != nil && Ops.greater_than(str_offset, 0)
          stringfirst = Builtins.substring(url, 0, str_offset)
          stringsecond = Builtins.substring(url, str_offset)

          Builtins.y2debug(
            "Pos: %1 First: >%2< Second: >%3<",
            str_offset,
            stringfirst,
            stringsecond
          )

          stringfirst = Builtins.regexpsub(stringfirst, "^(.*/)[^/]+/*$", "\\1")
          stringsecond = Builtins.regexpsub(
            stringsecond,
            "^/\\.\\./(.*)$",
            "\\1"
          )

          url = Ops.add(stringfirst, stringsecond)
        end
      end

      # remove /./
      max_count = 100

      while Ops.greater_than(max_count, 0) && Builtins.regexpmatch(url, "/\\./")
        max_count = Ops.subtract(max_count, 1)
        url = Builtins.regexpsub(url, "^(.*)/\\./(.*)", "\\1/\\2")
      end

      base_params_map = MakeMapFromParams(base_params)
      added_params_map = MakeMapFromParams(added_params)
      final_params_map = Convert.convert(
        Builtins.union(base_params_map, added_params_map),
        :from => "map",
        :to   => "map <string, string>"
      )

      if Ops.greater_than(Builtins.size(final_params_map), 0)
        Builtins.y2milestone(
          "%1 merge %2 -> %3",
          base_params_map,
          added_params_map,
          final_params_map
        )

        url_add = ""

        Builtins.foreach(final_params_map) do |opt, val|
          url_add = Ops.add(
            Ops.add(
              Ops.add(
                url_add,
                # add '&' if needed
                Ops.greater_than(Builtins.size(url_add), 0) ? "&" : ""
              ),
              opt
            ),
            # define 'opt=val' if val is set
            # otherwise don't use '=val' at all
            Ops.greater_than(Builtins.size(val), 0) ? Ops.add("=", val) : ""
          )
        end

        url = Ops.add(Ops.add(url, "?"), url_add)
      end

      Builtins.y2milestone("Final URL: '%1'", url)
      url
    end
    # <--


    # Adapts the inst-sys from the tarball
    # @param [String] filename string the filename with the tarball to use to the update
    # @return [Boolean] true on success
    def UpdateInstSys(filename)
      @src_cache_id = Ops.add(@src_cache_id, 1)
      tmpdir = Convert.to_string(SCR.Read(path(".target.tmpdir")))
      tmpdir = Builtins.sformat("%1/%2", tmpdir, @src_cache_id)
      out = Convert.to_map(
        SCR.Execute(
          path(".target.bash_output"),
          Builtins.sformat(
            "\n" +
              "/bin/mkdir %1;\n" +
              "cd %1;\n" +
              "/bin/tar -xvf %2;\n" +
              "/sbin/adddir %1 /;\n",
            tmpdir,
            filename
          )
        )
      )
      if Ops.get_integer(out, "exit", 0) != 0
        Builtins.y2error("Including installation image failed: %1", out)
        return false
      end
      Builtins.y2milestone("Including installation image succeeded")
      true
    end

    # Check all proposals, split those ones which have multiple modes or
    # architectures or stages into multiple proposals
    # @param [Array<Hash>] proposals a list of proposals
    # @return a list of updated proposals
    def PrepareProposals(proposals)
      proposals = deep_copy(proposals)
      new_proposals = []
      Builtins.foreach(proposals) do |p|
        mode = Ops.get_string(p, "mode", "")
        modes = Builtins.splitstring(mode, ",")
        modes = [""] if Builtins.size(modes) == 0
        Builtins.foreach(modes) do |m|
          mp = deep_copy(p)
          Ops.set(mp, "mode", m)
          arch = Ops.get_string(p, "archs", "")
          archs = Builtins.splitstring(arch, ",")
          archs = [""] if Builtins.size(archs) == 0
          Builtins.foreach(archs) do |a|
            amp = deep_copy(mp)
            Ops.set(amp, "archs", a)
            stage = Ops.get_string(amp, "stage", "")
            stages = Builtins.splitstring(stage, ",")
            stages = [""] if Builtins.size(stages) == 0
            Builtins.foreach(stages) do |s|
              samp = deep_copy(amp)
              Ops.set(samp, "stage", s)
              new_proposals = Builtins.add(new_proposals, samp)
            end
          end
        end
      end
      deep_copy(new_proposals)
    end

    # Check all proposals, split those ones which have multiple modes or
    # architectures or stages into multiple proposals.
    # Works with base product proposals.
    def PrepareSystemProposals
      return if @system_proposals_prepared
      ProductControl.proposals = PrepareProposals(ProductControl.proposals)
      @system_proposals_prepared = true

      nil
    end

    # Check all workflows, split those ones which have multiple modes or
    # architectures or stages into multiple workflows
    # @param [Array<Hash>] workflows a list of workflows
    # @return a list of updated workflows
    def PrepareWorkflows(workflows)
      workflows = deep_copy(workflows)
      new_workflows = []
      Builtins.foreach(workflows) do |w|
        mode = Ops.get_string(w, "mode", "")
        modes = Builtins.splitstring(mode, ",")
        modes = [""] if Builtins.size(modes) == 0
        Builtins.foreach(modes) do |m|
          mw = deep_copy(w)
          Ops.set(mw, "mode", m)
          Ops.set(mw, "defaults", Ops.get_map(mw, "defaults", {}))
          arch = Ops.get_string(mw, ["defaults", "archs"], "")
          archs = Builtins.splitstring(arch, ",")
          archs = [""] if Builtins.size(archs) == 0
          Builtins.foreach(archs) do |a|
            amw = deep_copy(mw)
            Ops.set(amw, ["defaults", "archs"], a)
            stage = Ops.get_string(amw, "stage", "")
            stages = Builtins.splitstring(stage, ",")
            stages = [""] if Builtins.size(stages) == 0
            Builtins.foreach(stages) do |s|
              samw = deep_copy(amw)
              Ops.set(samw, "stage", s)
              new_workflows = Builtins.add(new_workflows, samw)
            end
          end
        end
      end
      deep_copy(new_workflows)
    end

    # Check all workflows, split those ones which have multiple modes or
    # architectures or stages into multiple worlflows.
    # Works with base product workflows.
    def PrepareSystemWorkflows
      return if @system_workflows_prepared
      ProductControl.workflows = PrepareWorkflows(ProductControl.workflows)
      @system_workflows_prepared = true

      nil
    end

    # Stores the initial settings for workflows, proposals, inst_finish, clone_modules
    # to be able to start from scratch when some part of the workflow modified by add-on
    # is removed.
    def StoreInitialSettings
      # store it only once
      return if @initial_settings_stored

      Builtins.y2milestone("Storing initial settings")
      @initial_workflows = deep_copy(ProductControl.workflows)
      @initial_proposals = deep_copy(ProductControl.proposals)
      @initial_inst_finish = deep_copy(ProductControl.inst_finish)
      @initial_clone_modules = deep_copy(ProductControl.clone_modules)

      @initial_settings_stored = true

      nil
    end

    # Reset all the workflow and proposals modifications done
    # by Add-on products.
    # @see #StoreInitialSettings()
    def ResetSystemWorkflowsAndProposals
      Builtins.y2milestone("Restoring initial settings")
      ProductControl.workflows = deep_copy(@initial_workflows)
      ProductControl.proposals = deep_copy(@initial_proposals)
      ProductControl.inst_finish = deep_copy(@initial_inst_finish)
      ProductControl.clone_modules = deep_copy(@initial_clone_modules)

      @system_proposals_prepared = false
      PrepareSystemProposals()

      @system_workflows_prepared = false
      PrepareSystemWorkflows()

      nil
    end

    # Replace a module in a proposal with a set of other modules
    # @param [Hash] proposal a map describing the proposal
    # @param [String] old string the old item to be replaced
    # @param [Array<String>] new a list of items to be put into instead of the old one
    # @return a map with the updated proposal
    def ReplaceProposalModule(proposal, old, new)
      proposal = deep_copy(proposal)
      new = deep_copy(new)
      found = false
      modules = Builtins.maplist(Ops.get_list(proposal, "proposal_modules", [])) do |m|
        if Ops.is_string?(m) && Convert.to_string(m) == old ||
            Ops.is_map?(m) &&
              Ops.get_string(Convert.to_map(m), "name", "") == old
          found = true
          if Ops.is_map?(m)
            next Builtins.maplist(new) do |it|
              Builtins.union(Convert.to_map(m), { "name" => it })
            end
          else
            next deep_copy(new)
          end
        else
          next [m]
        end
      end
      if !found
        Builtins.y2internal("Replace/Remove proposal item %1 not found", old)
      end
      Ops.set(proposal, "proposal_modules", Builtins.flatten(modules))
      if Builtins.haskey(proposal, "proposal_tabs")
        Ops.set(
          proposal,
          "proposal_tabs",
          Builtins.maplist(Ops.get_list(proposal, "proposal_tabs", [])) do |tab|
            modules2 = Builtins.maplist(
              Ops.get_list(tab, "proposal_modules", [])
            ) do |m|
              if m == old
                next deep_copy(new)
              else
                next [m]
              end
            end
            Ops.set(tab, "proposal_modules", Builtins.flatten(modules2))
            deep_copy(tab)
          end
        )
      end
      deep_copy(proposal)
    end


    # Merge add-on proposal to a base proposal
    # @param [Hash] base map the base product proposal
    # @param [Hash] addon map the proposal of the addon productA
    # @param [String] prod_name a name of the add-on product
    # @return [Hash] merged proposals
    def MergeProposal(base, addon, prod_name, domain)
      base = deep_copy(base)
      addon = deep_copy(addon)
      appends = Ops.get_list(addon, "append_modules", [])
      removes = Ops.get_list(addon, "remove_modules", [])
      replaces = Builtins.listmap(Ops.get_list(addon, "replace_modules", [])) do |a|
        old = Ops.get_string(a, "replace", "")
        new = Ops.get_list(a, "modules", [])
        { old => new }
      end
      Builtins.foreach(removes) { |r| base = ReplaceProposalModule(base, r, []) } if Ops.greater_than(
        Builtins.size(removes),
        0
      )
      Builtins.foreach(replaces) do |old, new|
        base = ReplaceProposalModule(base, old, new)
      end if Ops.greater_than(
        Builtins.size(replaces),
        0
      )
      if Ops.greater_than(Builtins.size(appends), 0)
        as_map = false
        append2 = deep_copy(appends)
        if Ops.is_map?(Ops.get(base, ["proposal_modules", 0]))
          append2 = Builtins.maplist(appends) do |m|
            { "name" => m, "presentation_order" => 9999 }
          end
        end
        Ops.set(
          base,
          "proposal_modules",
          Builtins.merge(Ops.get_list(base, "proposal_modules", []), append2)
        )
        if Builtins.haskey(base, "proposal_tabs")
          new_tab = {
            "label"            => prod_name,
            "proposal_modules" => appends,
            "textdomain"       => domain
          }
          Ops.set(
            base,
            "proposal_tabs",
            Builtins.add(Ops.get_list(base, "proposal_tabs", []), new_tab)
          )
        end
      end
      if Ops.get_string(addon, "enable_skip", "yes") == "no"
        Ops.set(base, "enable_skip", "no")
      end
      deep_copy(base)
    end

    # Update system proposals according to proposal update metadata
    # @param [Array<Hash>] proposals a list of update proposals
    # @param [String] prod_name string the product name (used in case of tabs)
    # @param [String] domain string the text domain (for translations)
    # @return [Boolean] true on success
    def UpdateProposals(proposals, prod_name, domain)
      proposals = deep_copy(proposals)
      Builtins.foreach(proposals) do |proposal|
        name = Ops.get_string(proposal, "name", "")
        stage = Ops.get_string(proposal, "stage", "")
        mode = Ops.get_string(proposal, "mode", "")
        arch = Ops.get_string(proposal, "archs", "")
        found = false
        new_proposals = []
        arch_all_prop = {}
        Builtins.foreach(ProductControl.proposals) do |p|
          if Ops.get_string(p, "stage", "") != stage ||
              Ops.get_string(p, "mode", "") != mode ||
              Ops.get_string(p, "name", "") != name
            new_proposals = Builtins.add(new_proposals, p)
            next
          end
          if Ops.get_string(p, "archs", "") == arch || arch == "" ||
              arch == "all"
            p = MergeProposal(p, proposal, prod_name, domain)
            found = true
          elsif Ops.get_string(p, "archs", "") == "" ||
              Ops.get_string(p, "archs", "") == "all"
            arch_all_prop = deep_copy(p)
          end
          new_proposals = Builtins.add(new_proposals, p)
        end
        if !found
          if arch_all_prop != {}
            Ops.set(arch_all_prop, "archs", arch)
            proposal = MergeProposal(arch_all_prop, proposal, prod_name, domain) # completly new proposal
          else
            Ops.set(proposal, "textdomain", domain)
          end
          new_proposals = Builtins.add(new_proposals, proposal)
        end
        ProductControl.proposals = deep_copy(new_proposals)
      end
      true
    end


    # Replace a module in a workflow with a set of other modules
    # @param [Hash] workflow a map describing the workflow
    # @param [String] old string the old item to be replaced
    # @param [Array<Hash>] new a list of items to be put into instead of the old one
    # @param [String] domain string a text domain
    # @param [Boolean] keep boolean true to keep original one (and just insert before)
    # @return a map with the updated workflow
    def ReplaceWorkflowModule(workflow, old, new, domain, keep)
      workflow = deep_copy(workflow)
      new = deep_copy(new)
      found = false
      modules = Builtins.maplist(Ops.get_list(workflow, "modules", [])) do |m|
        if Ops.get_string(m, "name", "") == old
          new_list = Builtins.maplist(new) do |n|
            Ops.set(n, "textdomain", domain)
            deep_copy(n)
          end
          found = true
          new_list = Builtins.add(new_list, m) if keep
          next deep_copy(new_list)
        else
          next [m]
        end
      end
      if !found
        Builtins.y2internal(
          "Insert/Replace/Remove workflow module %1 not found",
          old
        )
      end
      Ops.set(workflow, "modules", Builtins.flatten(modules))
      deep_copy(workflow)
    end


    # Merge add-on workflow to a base workflow
    # @param [Hash] base map the base product workflow
    # @param [Hash] addon map the workflow of the addon product
    # @param [String] prod_name a name of the add-on product
    # @return [Hash] merged workflows
    def MergeWorkflow(base, addon, prod_name, domain)
      base = deep_copy(base)
      addon = deep_copy(addon)
      appends = Ops.get_list(addon, "append_modules", [])
      removes = Ops.get_list(addon, "remove_modules", [])
      inserts = Builtins.listmap(Ops.get_list(addon, "insert_modules", [])) do |i|
        before = Ops.get_string(i, "before", "")
        new = Ops.get_list(i, "modules", [])
        { before => new }
      end
      replaces = Builtins.listmap(Ops.get_list(addon, "replace_modules", [])) do |a|
        old = Ops.get_string(a, "replace", "")
        new = Ops.get_list(a, "modules", [])
        { old => new }
      end
      Builtins.foreach(removes) do |r|
        base = ReplaceWorkflowModule(base, r, [], domain, false)
      end if Ops.greater_than(
        Builtins.size(removes),
        0
      )
      Builtins.foreach(replaces) do |old, new|
        base = ReplaceWorkflowModule(base, old, new, domain, false)
      end if Ops.greater_than(
        Builtins.size(replaces),
        0
      )
      Builtins.foreach(inserts) do |old, new|
        base = ReplaceWorkflowModule(base, old, new, domain, true)
      end if Ops.greater_than(
        Builtins.size(inserts),
        0
      )
      Builtins.foreach(appends) do |new|
        Ops.set(new, "textdomain", domain)
        Ops.set(
          base,
          "modules",
          Builtins.add(Ops.get_list(base, "modules", []), new)
        )
      end if Ops.greater_than(
        Builtins.size(appends),
        0
      )
      deep_copy(base)
    end

    # Update system workflows according to workflow update metadata
    # @param [Array<Hash>] workflows a list of update workflows
    # @param [String] prod_name string the product name (used in case of tabs)
    # @param [String] domain string the text domain (for translations)
    # @return [Boolean] true on success
    def UpdateWorkflows(workflows, prod_name, domain)
      workflows = deep_copy(workflows)
      Builtins.foreach(workflows) do |workflow|
        stage = Ops.get_string(workflow, "stage", "")
        mode = Ops.get_string(workflow, "mode", "")
        arch = Ops.get_string(workflow, "archs", "")
        found = false
        new_workflows = []
        arch_all_wf = {}
        Builtins.foreach(ProductControl.workflows) do |w|
          if Ops.get_string(w, "stage", "") != stage ||
              Ops.get_string(w, "mode", "") != mode
            new_workflows = Builtins.add(new_workflows, w)
            next
          end
          if Ops.get_string(w, ["defaults", "archs"], "") == arch || arch == "" ||
              arch == "all"
            w = MergeWorkflow(w, workflow, prod_name, domain)
            found = true
          elsif Ops.get_string(w, ["defaults", "archs"], "") == "" ||
              Ops.get_string(w, ["default", "archs"], "") == "all"
            arch_all_wf = deep_copy(w)
          end
          new_workflows = Builtins.add(new_workflows, w)
        end
        if !found
          if arch_all_wf != {}
            Ops.set(arch_all_wf, ["defaults", "archs"], arch)
            workflow = MergeWorkflow(arch_all_wf, workflow, prod_name, domain) # completly new workflow
          else
            Ops.set(workflow, "textdomain", domain)
            Ops.set(
              workflow,
              "modules",
              Builtins.maplist(Ops.get_list(workflow, "modules", [])) do |mod|
                Ops.set(mod, "textdomain", domain)
                deep_copy(mod)
              end
            )
          end
          new_workflows = Builtins.add(new_workflows, workflow)
        end
        ProductControl.workflows = deep_copy(new_workflows)
      end
      true
    end

    # Redraw the wizard steps bar
    # @return [Boolean] true on success
    def RedrawWizardSteps
      # FIXME following code copy-pasted from clients/installation.ycp
      Builtins.y2milestone("Retranslating messages")

      # Make sure the labels for default function keys are retranslated, too.
      # Using Label::DefaultFunctionKeyMap() from Label module.
      UI.SetFunctionKeys(Label.DefaultFunctionKeyMap)

      # Activate language changes on static part of wizard dialog

      ProductControl.RetranslateWizardSteps
      Wizard.RetranslateButtons
      Wizard.SetFocusToNextButton
      # FIXME end of copy-paste
      true
    end

    # Add specified steps to inst_finish
    # Just modifies internal variables, inst_finish grabs them itself
    # @param [Hash{String => Array<String>}] additional_steps a map specifying the steps to be added
    # @return [Boolean] true on success
    def UpdateInstFinish(additional_steps)
      additional_steps = deep_copy(additional_steps)
      before_chroot = Ops.get(additional_steps, "before_chroot", [])
      after_chroot = Ops.get(additional_steps, "after_chroot", [])
      before_umount = Ops.get(additional_steps, "before_umount", [])
      @finish_steps_before_chroot = Convert.convert(
        Builtins.merge(@finish_steps_before_chroot, before_chroot),
        :from => "list",
        :to   => "list <string>"
      )
      @finish_steps_after_chroot = Convert.convert(
        Builtins.merge(@finish_steps_after_chroot, after_chroot),
        :from => "list",
        :to   => "list <string>"
      )
      @finish_steps_before_umount = Convert.convert(
        Builtins.merge(@finish_steps_before_umount, before_umount),
        :from => "list",
        :to   => "list <string>"
      )
      true
    end

    # Adapts the installation workflow according to specified XML file
    # @param [Hash] update_file a map containing the add-on product control file
    # @param [String] name string the name of the add-on product
    # @param [String] domain string the text domain for the add-on product
    # @return [Boolean] true on success
    def UpdateInstallation(update_file, name, domain)
      update_file = deep_copy(update_file)
      PrepareSystemProposals()
      PrepareSystemWorkflows()
      proposals = Ops.get_list(update_file, "proposals", [])
      proposals = PrepareProposals(proposals)
      UpdateProposals(proposals, name, domain)
      workflows = Ops.get_list(update_file, "workflows", [])
      workflows = PrepareWorkflows(workflows)
      UpdateWorkflows(workflows, name, domain)
      true
    end

    # Add new defined proposal to the list of system proposals
    # @param [Array<Hash>] proposals a lsit of proposals to be added
    # @return [Boolean] true on success
    def AddNewProposals(proposals)
      proposals = deep_copy(proposals)
      forbidden = Builtins.maplist(ProductControl.proposals) do |p|
        Ops.get_string(p, "name", "")
      end
      forbidden = Builtins.toset(forbidden)
      Builtins.foreach(proposals) do |proposal|
        if !Builtins.contains(forbidden, Ops.get_string(proposal, "name", ""))
          Builtins.y2milestone(
            "Adding new proposal %1",
            Ops.get_string(proposal, "name", "")
          )
          ProductControl.proposals = Builtins.add(
            ProductControl.proposals,
            proposal
          )
        end
      end
      true
    end

    # Replace workflows for 2nd stage of installation
    # @param [Array<Hash>] workflows a list of the workflows
    # @return [Boolean] true on success
    def Replaceworkflows(workflows)
      workflows = deep_copy(workflows)
      workflows = PrepareWorkflows(workflows)
      workflows = Builtins.filter(workflows) do |w|
        if Ops.get_string(w, "stage", "") == "initial"
          Builtins.y2error(
            "Attempting to replace 1st stage workflow. This is not possible"
          )
          next false
        end
        true
      end
      sm = {}
      Builtins.foreach(workflows) do |w|
        Ops.set(
          sm,
          Ops.get_string(w, "stage", ""),
          Ops.get(sm, Ops.get_string(w, "stage", ""), {})
        )
        Ops.set(
          sm,
          [Ops.get_string(w, "stage", ""), Ops.get_string(w, "mode", "")],
          true
        )
        [Ops.get_string(w, "stage", ""), Ops.get_string(w, "mode", "")]
      end

      Builtins.y2milestone("Existing replace workflows: %1", sm)
      Builtins.y2milestone(
        "Workflows before filtering: %1",
        Builtins.size(ProductControl.workflows)
      )
      ProductControl.workflows = Builtins.filter(ProductControl.workflows) do |w|
        !Ops.get(
          sm,
          [Ops.get_string(w, "stage", ""), Ops.get_string(w, "mode", "")],
          false
        )
      end
      Builtins.y2milestone(
        "Workflows after filtering: %1",
        Builtins.size(ProductControl.workflows)
      )
      ProductControl.workflows = Convert.convert(
        Builtins.merge(ProductControl.workflows, workflows),
        :from => "list",
        :to   => "list <map>"
      )

      true
    end

    # Update product options
    # @param [Hash] update_file a map containing update control file
    # @return [Boolean] true on success
    def UpdateProductInfo(update_file)
      update_file = deep_copy(update_file)
      Builtins.foreach(
        ["globals", "software", "partitioning", "network", "texts"]
      ) do |section|
        sect = ProductFeatures.GetSection(section)
        addon = Ops.get_map(update_file, section, {})
        sect = Convert.convert(
          Builtins.union(sect, addon),
          :from => "map",
          :to   => "map <string, any>"
        )
        ProductFeatures.SetSection(section, sect)
      end

      # merging 'clone_modules'
      addon_clone = Ops.get_list(update_file, "clone_modules", [])
      ProductControl.clone_modules = Convert.convert(
        Builtins.merge(ProductControl.clone_modules, addon_clone),
        :from => "list",
        :to   => "list <string>"
      )

      # merging texts

      #
      # **Structure:**
      #
      #     $[
      #        "congratulate" : $[
      #          "label" : "some text",
      #        ],
      #        "congratulate2" : $[
      #          "label" : "some other text",
      #          "textdomain" : "control-2", // (optionally)
      #        ],
      #      ];
      controlfile_texts = ProductFeatures.GetSection("texts")
      update_file_texts = Ops.get_map(update_file, "texts", {})
      update_file_textdomain = Ops.get_string(update_file, "textdomain", "")

      # if textdomain is different to the base one
      # we have to put it into the map
      if update_file_textdomain != nil && update_file_textdomain != ""
        update_file_texts = Builtins.mapmap(update_file_texts) do |text_ident, text_def|
          Ops.set(text_def, "textdomain", update_file_textdomain)
          { text_ident => text_def }
        end
      end

      controlfile_texts = Convert.convert(
        Builtins.union(controlfile_texts, update_file_texts),
        :from => "map",
        :to   => "map <string, any>"
      )
      ProductFeatures.SetSection("texts", controlfile_texts)

      true
    end

    # New add-on product might add also new agents.
    # Functions Rereads all available agents.
    #
    # @see bugzilla #239055, #245508
    def RereadAllSCRAgents
      Builtins.y2milestone("Registering new agents...")
      ret = SCR.RegisterNewAgents

      if ret
        Builtins.y2milestone("Successful")
      else
        Builtins.y2error("Error occured during registering new agents!")
        Report.Error(
          _("An error occurred while preparing the installation system.")
        )
      end

      nil
    end

    # Remove the /y2update directory from the system
    def CleanY2Update
      SCR.Execute(path(".target.bash"), "/bin/rm -rf /y2update")

      nil
    end

    # Show /media.1/info.txt file in a pop-up message if such file exists.
    # Show license if such exists and return whether users accepts it.
    # Returns 'nil' when did not succed.
    #
    # @return [Boolean] whether the license has been accepted
    def AcceptedLicenseAndInfoFile(src_id)
      ret = ProductLicense.AskAddOnLicenseAgreement(src_id)
      if ret == nil
        return nil
      elsif ret == :abort || ret == :back
        Builtins.y2milestone("License confirmation failed")
        return false
      end
      true
    end

    #symbol DoInstall_NoControlFile () {
    #    y2milestone("File /installation.xml not found, running sw_single for this source");
    #   WFM::CallFunction ("sw_single", []);

    #    return `register;
    #}

    def AnyPatternInRepo
      patterns = Pkg.ResolvableProperties("", :pattern, "")

      Builtins.y2milestone(
        "Total number of patterns: %1",
        Builtins.size(patterns)
      )

      patterns = Builtins.filter(patterns) do |pat|
        Ops.get(pat, "source") == @src_id
      end

      Builtins.y2milestone("Found %1 add-on patterns", Builtins.size(patterns))
      Builtins.y2debug("Found add-on patterns: %1", patterns)

      Ops.greater_than(Builtins.size(patterns), 0)
    end

    def DoInstall_NoControlFile
      Builtins.y2milestone(
        "File /installation.xml not found, running sw_single for this repository"
      )

      # display pattern the dialog when there is a pattern provided by the addon
      # otherwise use search mode
      cached_control_file = "/etc/opt/novell/oes-installation.xml"
      if !FileUtils.Exists(cached_control_file)
        Builtins.y2milestone(
          "Cached OES control file %1 not found",
          cached_control_file
        )
      end

      ProductControl.custom_control_file = cached_control_file
      if !ProductControl.Init
        # error report
        Report.Error(
          Builtins.sformat(_("Control file %1 not found "), cached_control_file)
        )
        return nil
      end

      # start workflow
      Wizard.OpenNextBackStepsDialog
      # dialog caption
      Wizard.SetContents(_("Initializing..."), Empty(), "", false, false)
      stage_mode = [{ "stage" => "normal", "mode" => "installation" }]
      ProductControl.AddWizardSteps(stage_mode)
      Mode.SetMode("installation")
      ret = ProductControl.Run
      UI.CloseDialog

      return :abort if ret == :abort || ret == :cancel || ret == :close

      nil
    end

    def DoInstall_WithControlFile(control)
      Builtins.y2milestone(
        "File /installation.xml was found, running own workflow..."
      )
      # copy the control file to local filesystem - in case of media release
      tmp = Convert.to_string(SCR.Read(path(".target.tmpdir")))
      tmp = Ops.add(tmp, "/installation.xml")
      SCR.Execute(
        path(".target.bash"),
        Builtins.sformat("/bin/cp %1 %2", control, tmp)
      )
      control = tmp

      # The following line is the only change between this file and the real OesAddOnProduct.ycp.
      binaries = nil # Pkg::SourceProvideOptionalFile (src_id, 1, "/y2update.tgz");
      # File /y2update.tgz exists
      if binaries != nil
        # Try to extract files from the archive
        out = Convert.to_map(
          SCR.Execute(
            path(".target.bash_output"),
            Builtins.sformat(
              "\n" +
                "test -d /y2update && rm -rf /y2update;\n" +
                "/bin/mkdir -p /y2update/all;\n" +
                "cd /y2update/all;\n" +
                "/bin/tar -xvf %1;\n" +
                "cd /y2update;\n" +
                "ln -s all/usr/share/YaST2/* .;\n" +
                "ln -s all/usr/lib/YaST2/* .;\n",
              binaries
            )
          )
        )

        # Failed
        if Ops.get_integer(out, "exit", 0) != 0
          # error report
          Report.Error(
            _("An error occurred while preparing the installation system.")
          )
          CleanY2Update()
          return nil
        else
          # bugzilla #239055
          RereadAllSCRAgents()
        end
      else
        Builtins.y2milestone("File /y2update.tgz not provided")
      end

      # set control file
      ProductControl.custom_control_file = control
      if !ProductControl.Init
        # error report
        Report.Error(
          Builtins.sformat(_("Control file %1 not found on media."), control)
        )
        CleanY2Update()
        return nil
      end

      # start workflow
      Wizard.OpenNextBackStepsDialog
      # dialog caption
      Wizard.SetContents(_("Initializing..."), Empty(), "", false, false)
      stage_mode = [{ "stage" => "normal", "mode" => "installation" }]
      ProductControl.AddWizardSteps(stage_mode)
      Mode.SetMode("installation")
      ret = ProductControl.Run
      UI.CloseDialog
      CleanY2Update()

      ret
    end

    def ClearRegistrationRequest(src_id)
      Builtins.y2milestone(
        "Clearing registration flag for source ID %1",
        src_id
      )
      if src_id != nil
        @addons_requesting_registration = Builtins.filter(
          @addons_requesting_registration
        ) { |one_source| one_source != src_id }
      end

      nil
    end

    # Returns whether registration is requested by at least one of
    # used Add-On products.
    #
    # @return [Boolean] if requested
    def ProcessRegistration
      force_registration = false

      # checking add-on products one by one
      Builtins.foreach(@add_on_products) do |prod|
        srcid = Ops.get_integer(prod, "media")
        if srcid != nil &&
            Builtins.contains(@addons_requesting_registration, srcid)
          force_registration = true
          raise Break
        end
      end

      Builtins.y2milestone("Requesting registration: %1", force_registration)
      force_registration
    end

    # Add-On product might have been added into products requesting
    # registration. This pruduct has been removed (during configuring
    # list of add-on products).
    def RemoveRegistrationFlag(src_id)
      # filtering out src_id
      @addons_requesting_registration = Builtins.filter(
        @addons_requesting_registration
      ) { |one_id| one_id != src_id }

      # removing cached file
      tmpdir = Ops.add(
        Convert.to_string(SCR.Read(path(".target.tmpdir"))),
        "/add-on-content-files/"
      )
      cachedfile = Builtins.sformat("%1content-%2", tmpdir, src_id)
      if FileUtils.Exists(cachedfile)
        Builtins.y2milestone("Removing cached file %1", cachedfile)
        SCR.Execute(path(".target.remove"), cachedfile)
      end

      nil
    end

    # Checks whether the content file of the add-on has a flag REGISTERPRODUCT
    # set to "true" or "yes". If it has, product is added into list of pruducts
    # that need registration. Cached content file is used if possible.
    #
    # @param integer source id
    def PrepareForRegistration(src_id)
      tmpdir = Ops.add(
        Convert.to_string(SCR.Read(path(".target.tmpdir"))),
        "/add-on-content-files/"
      )

      # create directory if doesn't exist
      if !FileUtils.Exists(tmpdir)
        run = Convert.to_integer(
          SCR.Execute(
            path(".target.bash"),
            Builtins.sformat("/bin/mkdir -p '%1'", tmpdir)
          )
        )
        if run != 0
          Builtins.y2error("Cannot create directory %1", tmpdir)
          return nil
        end
      end

      # use cached file if possible
      contentfile = Builtins.sformat("%1content-%2", tmpdir, src_id)
      if FileUtils.Exists(contentfile)
        Builtins.y2milestone("Using cached contentfile %1", contentfile)
      else
        Builtins.y2milestone("Checking contentfile from source")
        sourcefile = Pkg.SourceProvideOptionalFile(src_id, 1, "/content")
        if sourcefile == nil
          Builtins.y2warning("Cannot obtain content file!")
          return nil
        end
        # copying content file
        run = Convert.to_integer(
          SCR.Execute(
            path(".target.bash"),
            Builtins.sformat("/bin/cp '%1' '%2'", sourcefile, contentfile)
          )
        )
        if run != 0
          Builtins.y2error("Cannot copy %1 to %2", sourcefile, contentfile)
          return nil
        end
      end

      # registering agent for the current content file
      SCR.RegisterAgent(
        path(".addon.content"),
        term(
          :ag_ini,
          term(
            :IniAgent,
            contentfile,
            {
              "options"  => ["read_only", "global_values", "flat"],
              "comments" => ["^#.*", "^[ \t]*$"],
              "params"   => [
                {
                  "match" => [
                    "^[ \t]*([a-zA-Z0-9_.]+)[ \t]*(.*)[ \t]*$",
                    "%s %s"
                  ]
                }
              ]
            }
          )
        )
      )
      register_product = Convert.to_string(
        SCR.Read(path(".addon.content.REGISTERPRODUCT"))
      )
      SCR.UnregisterAgent(path(".addon.content"))

      # evaluating REGISTERPRODUCT flag, default (nil == false)
      Builtins.y2milestone(
        "RegisterProduct flag for source %1 is %2",
        src_id,
        register_product
      )
      if register_product == "yes" || register_product == "true"
        @addons_requesting_registration = Builtins.add(
          @addons_requesting_registration,
          src_id
        )
      end

      nil
    end

    # Calls registration client if needed.
    #
    # @param integer source id
    def RegisterAddOnProduct(src_id)
      if Builtins.contains(@addons_requesting_registration, src_id)
        Builtins.y2milestone("Source ID %1 requests registration", src_id)
        WFM.CallFunction("inst_suse_register", [])
      else
        Builtins.y2milestone("Source ID %1 doesn't need registration", src_id)
      end

      nil
    end

    # Do installation of the add-on product within an installed system
    # srcid is got via OesAddOnProduct::src_id
    #
    # @param string src_id
    # @return [Symbol] the result symbol from wizard sequencer
    def DoInstall
      # Display /media.1/info.txt if such file exists
      # Display license and wait for agreement
      # Not needed here, license already shown in the workflow
      # boolean license_ret = AcceptedLicenseAndInfoFile(src_id);
      # if (license_ret != true) {
      # 	y2milestone("Removing the current source ID %1", src_id);
      # 	Pkg::SourceDelete(src_id);
      # 	return nil;
      # }

      # FATE #301312
      PrepareForRegistration(@src_id)

      # FATE #302398: PATTERNS keyword in content file
      HandleProductPATTERNS(@src_id)

      ret = nil

      control = Pkg.SourceProvideOptionalFile(@src_id, 1, "/control.xml")
      # Fallback -- Source didn't provide needed controll file
      # Handling as it was an installation source
      if control == nil
        ret = DoInstall_NoControlFile()
      else
        ret = DoInstall_WithControlFile(control)
      end

      # To copy installation.xml from OES media to /etc/opt/novell/oes-installation.xml
      NovellUtils.CopyControlFile

      if ret != nil
        # registers Add-On product if requested
        RegisterAddOnProduct(@src_id) if !Stage.normal && ret != :abort
      end

      Builtins.y2milestone("Returning: %1", ret)
      ret
    end

    # Integrate the changes in the workflow
    # @param [String] filename string filename of the control file (local filename)
    # @return [Boolean] true on success
    def WFIntegrate(filename)
      Builtins.y2milestone("WFIntegrate %1", filename)

      update_file = XML.XMLToYCPFile(filename)
      name = Ops.get_string(update_file, "display_name", "")
      ret = UpdateInstallation(
        Ops.get_map(update_file, "update", {}),
        name,
        Ops.get_string(update_file, "textdomain", "control")
      )

      if !ret
        Builtins.y2error("Failed to udpate installation workflow")
        return false
      end

      if !UpdateProductInfo(update_file)
        Builtins.y2error("Failed to set product options")
        return false
      end

      if !AddNewProposals(Ops.get_list(update_file, "proposals", []))
        Builtins.y2error("Failed to add new proposals")
        return false
      end

      if !Replaceworkflows(Ops.get_list(update_file, "workflows", []))
        Builtins.y2error("Failed to replace workflows")
        return false
      end

      if !RedrawWizardSteps()
        Builtins.y2error("Redrawing the wizard steps failed")
        return false
      end

      if !UpdateInstFinish(
          Ops.get_map(update_file, ["update", "inst_finish"], {})
        )
        Builtins.y2error("Adding inst_finish steps failed")
        return false
      end

      true
    end

    def TmpdirForControlFiles
      Ops.add(
        Convert.to_string(SCR.Read(path(".target.tmpdir"))),
        "/control_files"
      )
    end

    def SelectProductPatterns(content_file, src_id)
      cmd = Builtins.sformat(
        "grep '^PATTERNS[\t ]+' '%1' | sed 's/^PATTERNS[\t ]+//'",
        String.Quote(content_file)
      )

      cmd_out = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))

      if Ops.get_integer(cmd_out, "exit", -1) != 0
        Builtins.y2error("Command: >%1< returned: %2", cmd, cmd_out)

        return false
      end

      patterns_to_select = Builtins.splitstring(
        Ops.get_string(cmd_out, "stdout", ""),
        "\t "
      )
      patterns_to_select = Builtins.filter(patterns_to_select) do |one_pattern|
        one_pattern != nil && one_pattern != ""
      end

      if Builtins.size(patterns_to_select) == 0
        Builtins.y2milestone(
          "Add-On doesn't have any required patterns (PATTERNS in content)"
        )
        return true
      end

      Builtins.y2milestone(
        "Add-On requires these PATTERNS: %1",
        patterns_to_select
      )
      # clear/set
      Ops.set(@patterns_preselected_by_addon, src_id, [])

      ret = true

      Builtins.foreach(patterns_to_select) do |one_pattern|
        pattern_properties = Pkg.ResolvableProperties(one_pattern, :pattern, "")
        already_selected = false
        Builtins.foreach(pattern_properties) do |one_pattern_found|
          patt_status = Ops.get_symbol(one_pattern_found, "status", :unknown)
          # patern is already selected
          if patt_status == :installed || patt_status == :selected
            already_selected = true
            raise Break
          end
        end
        if already_selected
          Builtins.y2milestone(
            "Pattern %1 is already installed/selected",
            one_pattern
          )
          next
        end
        if !Pkg.ResolvableInstall(one_pattern, :pattern)
          Builtins.y2error(
            "Cannot select pattern: %1, reason: %2",
            one_pattern,
            Pkg.LastError
          )
          ret = false
        else
          Ops.set(
            @patterns_preselected_by_addon,
            src_id,
            Builtins.add(
              Ops.get(@patterns_preselected_by_addon, src_id, []),
              one_pattern
            )
          )
        end
      end

      ret
    end


    def DeselectProductPatterns(src_id)
      patterns_to_deselect = Ops.get(@patterns_preselected_by_addon, src_id, [])

      if Builtins.size(patterns_to_deselect) == 0
        Builtins.y2milestone("There's no pattern to be deselected")
        return true
      end

      ret = true

      Builtins.foreach(patterns_to_deselect) do |one_pattern|
        if !Pkg.ResolvableNeutral(one_pattern, :pattern, true)
          Builtins.y2error(
            "Cannot deselect pattern: %1, reason: %2",
            one_pattern,
            Pkg.LastError
          )
          ret = false
        end
      end

      ret
    end
    def HandleProductPATTERNS(srcid)
      # FATE #302398: PATTERNS keyword in content file
      content_file = Pkg.SourceProvideOptionalFile(srcid, 1, "/content")

      if content_file == nil
        Builtins.y2warning("Add-On %1 doesn't have a content file", srcid)
      else
        SelectProductPatterns(content_file, srcid)
      end

      nil
    end

    # Integrate the add-on product to the installation workflow, including
    # preparations for 2nd stage and inst-sys update
    # @param [Fixnum] srcid integer the source ID of the installation source
    # @return [Boolean] true on success
    def Integrate(srcid)
      Builtins.y2milestone("Integrating source %1", srcid)

      # bugzilla #297752
      # Update inst-sys before merging the control file

      # Updating inst-sys
      y2update = Pkg.SourceProvideOptionalFile(srcid, 1, "/y2update.tgz")
      if y2update == nil
        Builtins.y2milestone("No YaST update found on the media")
      else
        UpdateInstSys(y2update)
      end

      # FATE #302398: PATTERNS keyword in content file
      HandleProductPATTERNS(srcid)

      # Updating workflow or proposals

      # Trying to use a cached file, bugzilla #221377
      cached_filename = Builtins.sformat(
        "%1/%2.xml",
        TmpdirForControlFiles(),
        srcid
      )
      source_filename = nil

      # A cached copy exists
      if FileUtils.Exists(cached_filename)
        Builtins.y2milestone("Using cached file %1", cached_filename)
        source_filename = cached_filename 
        # Trying file from source
      else
        Builtins.y2milestone("File %1 not cached", cached_filename)
        # using a file from source
        source_filename = Pkg.SourceProvideOptionalFile(
          srcid,
          1,
          "/installation.xml"
        )
      end

      if source_filename == nil
        Builtins.y2milestone(
          "Add-on product control file not found, not touching the work-flow"
        ) 
        # either cached or file from source exists
      else
        # already cached
        if source_filename != cached_filename
          Builtins.y2milestone(
            "Copying %1 to %2",
            source_filename,
            cached_filename
          )
          tmpdir = TmpdirForControlFiles()
          out = Convert.to_map(
            SCR.Execute(
              path(".target.bash_output"),
              Builtins.sformat(
                "\n" +
                  "test -d '%1' || /bin/mkdir -p '%1';\n" +
                  "/bin/cp -v '%2' '%3';\n",
                tmpdir,
                source_filename,
                cached_filename
              )
            )
          )

          if Ops.get_integer(out, "exit", -1) != 0
            Builtins.y2error(
              "Error occurred while copying control file: %1",
              out
            )

            # cannot copy, using the non-cached file (fallback)
            cached_filename = source_filename
          end
        end

        # always use cached file
        if !WFIntegrate(cached_filename)
          Builtins.y2error("Workflow update failed")
          return false
        end

        @control_files_to_add = Builtins.add(
          @control_files_to_add,
          Builtins.sformat("%1.xml", srcid)
        )
      end

      true
    end

    # Removes (from disk cache) a control file which was included in
    # the add-on product.
    def RemoveAddOnControlFile(srcid)
      filename = Builtins.sformat("%1/%2.xml", TmpdirForControlFiles(), srcid)

      # remove also from list of control files
      # #238307
      this_control_file = Builtins.sformat("%1.xml", srcid)
      @control_files_to_add = Builtins.filter(@control_files_to_add) do |one_control_file|
        one_control_file != this_control_file
      end
      Builtins.y2milestone(
        "Control files in use now: %1",
        @control_files_to_add
      )

      if FileUtils.Exists(filename)
        Builtins.y2milestone("Removing: %1", filename)
        if Convert.to_boolean(SCR.Execute(path(".target.remove"), filename))
          return true
        else
          Builtins.y2error("Cannot remove file %1", filename)
          return false
        end
      else
        Builtins.y2milestone(
          "File %1 does not exist, no need to remove it",
          filename
        )
        return true
      end
    end

    # Some product(s) were removed, reintegrating their control files from scratch.
    def ReIntegrateFromScratch
      Builtins.y2milestone("Reintegration workflows from scratch...")
      ResetSystemWorkflowsAndProposals()
      # bugzilla #239055
      RereadAllSCRAgents()

      Builtins.foreach(@add_on_products) do |prod|
        srcid = Ops.get_integer(prod, "media")
        if srcid == nil
          Builtins.y2error(
            "Wrong definition of Add-on product: %1, cannot reintegrate",
            srcid
          )
          next
        else
          Builtins.y2milestone("Reintegrating product %1", prod)
          Integrate(srcid)
        end
      end

      # steps might be changed, forcing redraw
      Builtins.y2milestone("Forcing RedrawWizardSteps()")
      RedrawWizardSteps()

      true
    end

    def CheckProductDependencies(products)
      products = deep_copy(products)
      # TODO check the dependencies of the product
      true
    end

    # Auto-integrate add-on products in specified file
    # @param [String] filelist string a file containing a list of add-on products to integrate
    # @return [Boolean] true on exit
    def AddPreselectedAddOnProducts(filelist)
      if filelist == nil
        Builtins.y2milestone("No add-on products defined on the media")
        return true
      end
      products = Builtins.splitstring(
        Convert.to_string(SCR.Read(path(".target.string"), filelist)),
        " \n"
      )

      base_url = GetBaseProductURL()
      Builtins.y2milestone("Base URL: %1", base_url)

      Builtins.y2milestone("Adding products: %1", products)
      Builtins.foreach(products) do |p|
        next if p == ""
        elements = Builtins.splitstring(p, " \t")
        elements = Builtins.filter(elements) { |e| e != "" }
        url = Ops.get(elements, 0, "")
        pth = Ops.get(elements, 1, "/")
        elements = Builtins.remove(elements, 0) if Ops.get(elements, 0) != nil
        elements = Builtins.remove(elements, 0) if Ops.get(elements, 0) != nil
        # FATE #302123
        url = GetAbsoluteURL(base_url, url) if base_url != nil && base_url != ""
        Builtins.y2milestone("Adding Repository: %1 %2", url, pth)
        src = Pkg.SourceCreate(url, pth)
        if !AcceptedLicenseAndInfoFile(src)
          Pkg.SourceDelete(src)
          next
        end
        Integrate(src)
        # adding the product to the list of products
        # bugzilla #269625
        prod = Convert.convert(
          Pkg.SourceProductData(src),
          :from => "map <string, any>",
          :to   => "map <string, string>"
        )
        @add_on_products = Builtins.add(
          @add_on_products,
          {
            "media"            => src,
            "product"          => Ops.get(
              prod,
              "label",
              Ops.get(prod, "productname", Ops.get(prod, "productversion", ""))
            ),
            "autoyast_product" => Ops.get(prod, "productname", ""),
            "media_url"        => url,
            "product_dir"      => pth
          }
        )
        if Ops.greater_than(Builtins.size(elements), 0)
          Builtins.foreach(elements) { |e| Pkg.ResolvableInstall(e, :product) }
        else
          products2 = Pkg.ResolvableProperties("", :product, "")
          products2 = Builtins.filter(products2) do |p2|
            Ops.get_integer(p2, "source", -1) == src
          end
          Builtins.foreach(products2) do |p2|
            Pkg.ResolvableInstall(Ops.get_string(p2, "name", ""), :product)
          end
        end
      end

      nil
    end

    def Export
      # See bugzilla #276309

      # Returns map like this one
      #
      # <add-on>
      #   <add_on_products config:type="list">
      #     <listentry>
      #       <media_url>ftp://server.name/.../</media_url>
      #       <product>NEEDS_TO_MATCH_"PRODUCT"_TAG_FROM_content_FILE!</product>
      #       <product_dir>/</product_dir>
      #     </listentry>
      #     ...
      #   </add_on_products>
      # </add-on>

      # Just to make sure everything is initialized properly
      Pkg.TargetInitialize(Installation.destdir)
      Pkg.TargetLoad

      # Get list of currently installed products
      products = Pkg.ResolvableProperties("", :product, "")

      exp = Builtins.maplist(@add_on_products) do |p|
        p = Builtins.remove(p, "media") if Builtins.haskey(p, "media")
        Builtins.y2milestone("Fixing product: %1", p)
        # Try to match the product "display_name" or "short_name"
        # with "product" of the exported Add-On
        Builtins.foreach(products) do |prod|
          Builtins.y2milestone("Trying... Product resolvable: %1", prod)
          if Ops.get_string(prod, "display_name", "") ==
              Ops.get_string(p, "product", "___no_product_name_") ||
              Ops.get_string(prod, "short_name", "") ==
                Ops.get_string(p, "product", "___no_product_name_")
            Builtins.y2milestone("Passed")
            Ops.set(p, "product", Ops.get_string(prod, "name", ""))
            # We do not need further testing for this product
            raise Break
          end
        end
        Builtins.y2milestone("Product to save: %1", p)
        deep_copy(p)
      end

      { "add_on_products" => exp }
    end

    def Import(settings)
      settings = deep_copy(settings)
      @add_on_products = Ops.get_list(settings, "add_on_products", [])
      @modified = false
      Builtins.foreach(@add_on_products) do |prod|
        media = Ops.get_string(prod, "media_url", "")
        pth = Ops.get_string(prod, "product_dir", "/")
        src = Pkg.SourceCreate(media, pth)
        if src != -1
          @mode_config_sources = Builtins.add(@mode_config_sources, src)
        end
      end if Mode.config(
      )
      true
    end

    def CleanModeConfigSources
      Builtins.foreach(@mode_config_sources) { |src| Pkg.SourceDelete(src) }
      @mode_config_sources = []

      nil
    end

    # Returns the path where Add-Ons configuration is stored during the fist stage installation.
    # This path reffers to the installed system.
    #
    # @see bugzilla #187558
    def TmpExportFilename
      Ops.add(Directory.vardir, "/exported_add_ons_configuration")
    end

    # Reads the Add-Ons configuration stored on disk during the first stage installation.
    #
    # @see bugzilla #187558
    def ReadTmpExportFilename
      tmp_filename = TmpExportFilename()
      @modified = true

      if FileUtils.Exists(tmp_filename)
        Builtins.y2milestone("Reading %1 content", tmp_filename)

        # there might be something already set, store the current configuration
        already_in_configuration = deep_copy(@add_on_products)
        configuration_from_disk = Convert.to_map(
          SCR.Read(path(".target.ycp"), tmp_filename)
        )
        Builtins.y2milestone(
          "Configuration from disk: %1",
          configuration_from_disk
        )

        if configuration_from_disk != nil
          Import(configuration_from_disk)
          if already_in_configuration != [] && already_in_configuration != nil
            @add_on_products = Convert.convert(
              Builtins.union(@add_on_products, already_in_configuration),
              :from => "list",
              :to   => "list <map <string, any>>"
            )
          end
          return true
        else
          Builtins.y2error("Reading %1 file returned nil result!", tmp_filename)
          return false
        end
      else
        Builtins.y2warning("File %1 doesn't exists, skipping...", tmp_filename)
        return true
      end
    end

    def Disintegrate(srcid)
      DeselectProductPatterns(srcid)

      nil
    end

    publish :variable => :add_on_products, :type => "list <map <string, any>>"
    publish :variable => :src_id, :type => "integer"
    publish :variable => :last_ret, :type => "symbol"
    publish :variable => :control_files_to_add, :type => "list <string>"
    publish :variable => :finish_steps_before_chroot, :type => "list <string>"
    publish :variable => :finish_steps_after_chroot, :type => "list <string>"
    publish :variable => :finish_steps_before_umount, :type => "list <string>"
    publish :variable => :modified, :type => "boolean"
    publish :variable => :mode_config_sources, :type => "list <integer>"
    publish :function => :RedrawWizardSteps, :type => "boolean ()"
    publish :function => :SetBaseProductURL, :type => "void (string)"
    publish :function => :GetBaseProductURL, :type => "string ()"
    publish :function => :GetAbsoluteURL, :type => "string (string, string)"
    publish :function => :UpdateInstSys, :type => "boolean (string)"
    publish :function => :StoreInitialSettings, :type => "void ()"
    publish :function => :RereadAllSCRAgents, :type => "void ()"
    publish :function => :AcceptedLicenseAndInfoFile, :type => "boolean (integer)"
    publish :function => :DoInstall_NoControlFile, :type => "symbol ()"
    publish :function => :ClearRegistrationRequest, :type => "void (integer)"
    publish :function => :ProcessRegistration, :type => "boolean ()"
    publish :function => :RemoveRegistrationFlag, :type => "void (integer)"
    publish :function => :PrepareForRegistration, :type => "void (integer)"
    publish :function => :RegisterAddOnProduct, :type => "void (integer)"
    publish :function => :DoInstall, :type => "symbol ()"
    publish :function => :WFIntegrate, :type => "boolean (string)"
    publish :function => :Integrate, :type => "boolean (integer)"
    publish :function => :RemoveAddOnControlFile, :type => "boolean (integer)"
    publish :function => :ReIntegrateFromScratch, :type => "boolean ()"
    publish :function => :CheckProductDependencies, :type => "boolean (list <string>)"
    publish :function => :AddPreselectedAddOnProducts, :type => "boolean (string)"
    publish :function => :Export, :type => "map ()"
    publish :function => :Import, :type => "boolean (map)"
    publish :function => :CleanModeConfigSources, :type => "void ()"
    publish :function => :TmpExportFilename, :type => "string ()"
    publish :function => :ReadTmpExportFilename, :type => "boolean ()"
    publish :function => :Disintegrate, :type => "void (integer)"
  end

  OesAddOnProduct = OesAddOnProductClass.new
  OesAddOnProduct.main
end
