#! /usr/bin/perl -w

# ------------------------------------------------------------------------------
# Copyright (c) 2006 Novell, Inc. All Rights Reserved.
#
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of version 2 of the GNU General Public License as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, contact Novell, Inc.
#
# To contact Novell about this file by physical or electronic mail, you may find
# current contact information at www.novell.com.
# ------------------------------------------------------------------------------

# File:         modules/ValidationHelpers.pm
# Package:      novell-common
# Summary:      Novell validation helper routines
# Authors:      Ryan Partridge <rpartridge@novell.com>
#
# $Id: ValidationHelpers.pm 27914 2006-02-13 14:32:08Z locilka $
#

package ValidationHelpers;

use strict;

use YaST::YCP qw(Boolean :LOGGING sformat);
use YaPI;
use Net::LDAP;
use Net::LDAP::Util qw (canonical_dn ldap_explode_dn escape_dn_value);

textdomain("novell-base");

our %TYPEINFO;

my $DEBUG = 0;

##
 # Convert a DN to LDAP format
 # @param dn The Distinguished Name to convert
 # @return string The converted DN, or "" if the DN was invalid
BEGIN { $TYPEINFO{toLdap} = ["function", "string", "string"]; }
sub toLdap
{
  my $self = shift;
  my ($rdn) = @_;
  my $ret = "";

#  print "RDN to validate: " . $rdn . "\n" if $DEBUG;

  # check for ambiguity
 if (defined $rdn && $rdn ne "")
 {
  if ($rdn =~ m/\.,/ || $rdn =~ m/,\./)
  {
    print "either ., or ,. found -- ambiguous format\n" if $DEBUG;
  }
  else
  {
    my $exploded = ldap_explode_dn($rdn, casefold => 'lower');
    # now put the RDN back together
    foreach my $part (@$exploded)
    {
      foreach my $type (keys(%$part))
      {
        print $type . ": " . $$part{$type} . "\n" if $DEBUG;
        $ret = $ret . "," . $type . "=" . $$part{$type};
      }
    }
    if (length($ret) > 0)
    {
      $ret = substr($ret, 1);
    }

    if ($rdn eq $ret)
    {
      # may be in NDS format

      # replace \. with |
      $rdn =~ s/\\\./|/g;
      print "(modified) " . $rdn . "\n" if $DEBUG;

      # replace \, with ,
      $rdn =~ s/\\\,/,/g;
      print "(modified) " . $rdn . "\n" if $DEBUG;

      # replace , with \,
      $rdn =~ s/,/\\\,/g;
      print "(modified) " . $rdn . "\n" if $DEBUG;

      # replace . with ,
      $rdn =~ s/\./,/g;
      print "(modified) " . $rdn . "\n" if $DEBUG;

      # replace | with .
      $rdn =~ s/\|/./g;
      print "(modified) " . $rdn . "\n" if $DEBUG;

      # try to canonicalize again
      $exploded = ldap_explode_dn($rdn, casefold => 'lower');

      $ret = "" ;
      # now put the RDN back together
      foreach my $part (@$exploded)
      {
        foreach my $type (keys(%$part))
        {
          print $type . ": " . $$part{$type} . "\n" if $DEBUG;
          $ret = $ret . "," . $type . "=" . $$part{$type};
        }
      }
      if (length($ret) > 0)
      {
        $ret = substr($ret, 1);
      }
    }

  }
 }
  return $ret;
}
##
 # escape the '.' for extend schema util
 # @param dn The Distinguished Name to convert
 # @return string The converted DN, or "" if the DN was invalid
BEGIN { $TYPEINFO{esdot} = ["function", "string", "string"]; }
sub esdot
{
  my $self = shift;
  my ($rdn) = @_;
  $rdn =~ s/\./\\\./g;
  print "(modified) " . $rdn . "\n" if $DEBUG;
  return $rdn
}



##
 # Convert a DN to NDS format
 # @param dn The Distinguished Name to convert
 # @return string The converted DN, or "" if the DN was invalid
BEGIN { $TYPEINFO{toNds} = ["function", "string", "string"]; }
sub toNds
{
  my $self = shift;
  my ($rdn) = @_;
  my $ret = "";

  # first convert to LDAP format (it's easier to validate)
  $rdn = $self->toLdap($rdn);
  if (defined $rdn && $rdn ne "")
  {
    # now convert from known good LDAP format to NDS format
    # replace \, with |
    $rdn =~ s/\\\,/|/g;
    print "(modified) " . $rdn . "\n" if $DEBUG;

    # replace . with \.
    $rdn =~ s/\./\\\./g;
    print "(modified) " . $rdn . "\n" if $DEBUG;

    # replace , with .
    $rdn =~ s/,/./g;
    print "(modified) " . $rdn . "\n" if $DEBUG;

    # replace | with ,
    $rdn =~ s/\|/,/g;
    print "(modified) " . $rdn . "\n" if $DEBUG;

    $ret = $rdn;
  }
  return $ret;
}

##
 # Validate an RDN context
 # @param context The context to validate
 # @return boolean True if the context is valid
BEGIN { $TYPEINFO{ValidateContext} = ["function", "boolean", "string"]; }
sub ValidateContext
{
  my $self = shift;
  my ($context) = @_;
  my %containment_rules = (
    c => ["t", "dc"],
    l => ["c", "ou", "l", "o", "dc", "s", "st"],
    s => ["l", "o", "c"],
    st => ["l", "o", "c"],
    o => ["t", "c", "l", "dc", "s", "st"],
    ou => ["l", "o", "ou", "dc", "cn", "s", "st"],
    dc => ["t", "c", "s", "st", "l", "o", "ou", "dc"],
    cn => ["cn", "o", "ou", "dc"],
  );
  my $ret = 1;
  my $has_o = 0;
  $context = $self->toLdap($context);
  print "Context: $context\n" if $DEBUG;
  if (defined $context && $context ne "")
  {
    my @types = ();
    my $exploded = ldap_explode_dn($context, casefold => 'lower');

    # build type list
    foreach my $rdn (@$exploded)
    {
      foreach my $type (keys(%$rdn))
      {
        print "Type: $type\n" if $DEBUG;
        push @types, $type;
      }
    }

    # validate based on containment rules
    for (my $i = 0; $i < scalar @types; ++$i)
    {
      my $containment_list_ref = [];

      # check the containment rules hash for this type
      # if the type is unknown, do not validate
      if (!exists($containment_rules{$types[$i]}))
      {
        print "unrecognized type ($types[$i])\n" if $DEBUG;
        $ret = 0;
        last;
      }
      else
      {
        print "checking type: $types[$i], contained by " if $DEBUG;
        $containment_list_ref = $containment_rules{$types[$i]};
      }
      
      # special case for the last element
      if ($i == $#types)
      {
        if (grep(/st/, @$containment_list_ref) == 1)
        {
           # Get rid of the "st" so it does not false trigger the check for "t"
           splice( @$containment_list_ref, -1);
        }
        if (grep(/t/, @$containment_list_ref) == 0)
        {
          print "$types[$i] may not be contained by the tree root!\n" if $DEBUG;
          $ret = 0;
          last;
        }
      }

      # otherwise look to the next element for proper containment
      else
      {
        my $container = $types[$i + 1];
        print "$container\n" if $DEBUG;
        if (grep(m/^$container$/, @$containment_list_ref) == 0)
        {
          print "$types[$i] may not be contained by $types[$i + 1]!\n" if $DEBUG;
          $ret = 0;
          last;
        }
      }
    }
  }
  else
  {
    $ret = 0;
  }
  return $ret;
}

##
 # Validate an FDN
 # @param fdn The context to validate
 # @return boolean True if the fdn is valid
BEGIN { $TYPEINFO{ValidateFDN} = ["function", "boolean", "string"]; }
sub ValidateFDN
{
  my $self = shift;
  my ($fdn) = @_;
  my $ret = 0;
  $fdn = $self->toLdap($fdn);
  if (defined $fdn && $fdn ne "")
  {
    if (substr($fdn, 0, 3) eq 'cn=')
    {
      print "FDN: $fdn\n" if $DEBUG;
      $fdn =~ s/^cn=.+?,//;
      print "Context for FDN: $fdn\n" if $DEBUG;
      if ($self->ValidateContext($fdn))
      {
        $ret = 1;
      }
    }
  }
  return $ret;
}

##
 # Validate credentials against given LDAP server
 # @param dn The DN of the user to validate
 # @param password The password to use in validation
 # @param host The address (IP/DNS) of the target server
 # @param port The LDAP port to use
 # @return boolean True if the credentials are valid
BEGIN { $TYPEINFO{ValidateCredentials} = ["function", "string", "string", "string", "string", "integer"]; }
sub ValidateCredentials
{
  my $self = shift;
  my ($dn, $password, $host, $port) = @_;
  my $ret = 0;
  my $mylang = "";

  $mylang = $ENV{'LANGUAGE'} if $ENV{'LANGUAGE'};
  y2milestone("ValidationHelpers::ValidateCredentials LANGUAGE=", $mylang);
  $dn = $self->toLdap($dn);
  my $ldap = Net::LDAP->new("ldaps://$host:$port");
  if (defined $ldap)
  {
    my $result = $ldap->bind($dn, password => $password);
    if ($result->code == 0)
    {
      $ret = 1;
      $ldap->unbind();
      $ldap->disconnect();
    }
    else
    {
    $ret= $result->error();
      y2milestone("LDAP bind error: ", $result->error());
    }
  }
  else
  {
    y2milestone("LDAP connect error: ", $@);
  }
  if (($mylang ne ""))
  {
    $ENV{'LANGUAGE'}=$mylang;
    y2milestone("ValidationHelpers::ValidateCredentials setting LANGUAGE=", $mylang);
  }

  else
  {
    delete @ENV{qw(LANGUAGE)} if $ENV{'LANGUAGE'};
    y2milestone("ValidationHelpers::ValidateCredentials  mylang was empty. LANGUAGE=", $mylang);
  }
  return $ret;
}

##
 # Validate credentials against given LDAP server
 # @param dn The DN of the user to validate
 # @param password The password to use in validation
 # @param host The address (IP/DNS) of the target server
 # @param port The LDAP port to use
 # @return boolean True if the credentials are valid
BEGIN { $TYPEINFO{ValidateCredentialsSamba} = ["function", "boolean", "string", "string", "string", "integer"]; }
sub ValidateCredentialsSamba
{
  my $self = shift;
  my ($dn, $password, $host, $port) = @_;
  my $ret = 0;
  my $mylang = "";

  $mylang = $ENV{'LANGUAGE'} if $ENV{'LANGUAGE'};
  y2milestone("ValidationHelpers::ValidateCredentials LANGUAGE=", $mylang);
  $dn = $self->toLdap($dn);
  my $ldap = Net::LDAP->new("ldaps://$host:$port");
  if (defined $ldap)
  {
    my $result = $ldap->bind($dn, password => $password);
    if ($result->code == 0)
    {
      $ret = 1;
      $ldap->unbind();
      $ldap->disconnect();
    }
    else
    {
      y2milestone("LDAP bind error: ", $result->error());
    }
  }
  else
  {
    y2milestone("LDAP connect error: ", $@);
  }
  if (($mylang ne ""))
  {
    $ENV{'LANGUAGE'}=$mylang;
    y2milestone("ValidationHelpers::ValidateCredentials setting LANGUAGE=", $mylang);
  }
  else
  {
    delete @ENV{qw(LANGUAGE)} if $ENV{'LANGUAGE'};
    y2milestone("ValidationHelpers::ValidateCredentials  mylang was empty. LANGUAGE=", $mylang);
  }
  return $ret;
}

1;
# EOF
